import os
from copy import deepcopy

import numpy as np
import pytest
import torch
from pythae.models.base.base_model import BaseAEConfig
from pythae.models.base.base_utils import ModelOutput

from multivae.data.datasets.base import IncompleteDataset, MultimodalBaseDataset
from multivae.models.auto_model.auto_model import AutoModel
from multivae.models.mvtcae import MVTCAE, MVTCAEConfig
from multivae.models.nn.default_architectures import Decoder_AE_MLP, Encoder_VAE_MLP
from multivae.trainers.base.base_trainer import BaseTrainer
from multivae.trainers.base.base_trainer_config import BaseTrainerConfig


class Test_model:
    @pytest.fixture(params=["complete", "incomplete"])
    def dataset(self, request):
        # Create simple small dataset
        data = dict(
            mod1=torch.randn((6, 2)),
            mod2=torch.randn((6, 3)),
            mod3=torch.randn((6, 4)),
            mod4=torch.randn((6, 4)),
        )
        labels = np.array([0, 1, 0, 0])
        if request.param == "complete":
            dataset = MultimodalBaseDataset(data, labels)
        else:
            masks = dict(
                mod1=torch.Tensor([True] * 3 + [False] * 3),
                mod2=torch.Tensor([True] * 6),
                mod3=torch.Tensor([True] * 6),
                mod4=torch.Tensor([True] * 6),
            )
            dataset = IncompleteDataset(data=data, masks=masks, labels=labels)

        return dataset

    @pytest.fixture
    def custom_architectures(self):
        # Create an instance of mvae model
        config1 = BaseAEConfig(input_dim=(2,), latent_dim=5)
        config2 = BaseAEConfig(input_dim=(3,), latent_dim=5)
        config3 = BaseAEConfig(input_dim=(4,), latent_dim=5)

        encoders = dict(
            mod1=Encoder_VAE_MLP(config1),
            mod2=Encoder_VAE_MLP(config2),
            mod3=Encoder_VAE_MLP(config3),
            mod4=Encoder_VAE_MLP(config3),
        )

        decoders = dict(
            mod1=Decoder_AE_MLP(config1),
            mod2=Decoder_AE_MLP(config2),
            mod3=Decoder_AE_MLP(config3),
            mod4=Decoder_AE_MLP(config3),
        )

        return dict(
            encoders=encoders,
            decoders=decoders,
        )

    @pytest.fixture(params=[1.0, 1.5, 2.0])
    def model_config(self, request):
        model_config = MVTCAEConfig(
            n_modalities=4,
            latent_dim=5,
            input_dims=dict(mod1=(2,), mod2=(3,), mod3=(4,), mod4=(4,)),
            beta=request.param,
        )

        return model_config

    @pytest.fixture(params=[True, False])
    def model(self, custom_architectures, model_config, request):
        custom = request.param
        if custom:
            model = MVTCAE(model_config, **custom_architectures)
        else:
            model = MVTCAE(model_config)
        return model

    def test(self, model, dataset, model_config):
        assert model.beta == model_config.beta

        output = model(dataset, epoch=2)
        loss = output.loss
        assert type(loss) == torch.Tensor
        assert loss.size() == torch.Size([])
        assert loss.requires_grad

        # Try encoding and prediction
        outputs = model.encode(dataset[0])
        assert outputs.one_latent_space
        embeddings = outputs.z
        assert isinstance(outputs, ModelOutput)
        assert embeddings.shape == (1, 5)
        embeddings = model.encode(dataset[0], N=2).z
        assert embeddings.shape == (2, 1, 5)
        embeddings = model.encode(dataset, cond_mod=["mod2"]).z
        assert embeddings.shape == (len(dataset), 5)
        embeddings = model.encode(dataset, cond_mod="mod3", N=10).z
        assert embeddings.shape == (10, len(dataset), 5)
        embeddings = model.encode(dataset, cond_mod=["mod2", "mod4"]).z
        assert embeddings.shape == (len(dataset), 5)

        Y = model.predict(dataset, cond_mod="mod2")
        assert isinstance(Y, ModelOutput)
        assert Y.mod1.shape == (len(dataset), 2)
        assert Y.mod2.shape == (len(dataset), 3)

        Y = model.predict(dataset, cond_mod="mod2", N=10)
        assert isinstance(Y, ModelOutput)
        assert Y.mod1.shape == (10, len(dataset), 2)
        assert Y.mod2.shape == (10, len(dataset), 3)

        Y = model.predict(dataset, cond_mod="mod2", N=10, flatten=True)
        assert isinstance(Y, ModelOutput)
        assert Y.mod1.shape == (len(dataset) * 10, 2)
        assert Y.mod2.shape == (len(dataset) * 10, 3)


class Test_backward_with_missing_inputs:
    @pytest.fixture(params=["incomplete"])
    def dataset(self, request):
        # Create simple small dataset
        data = dict(
            mod1=torch.randn((6, 2)),
            mod2=torch.randn((6, 3)),
            mod3=torch.randn((6, 4)),
            mod4=torch.randn((6, 4)),
        )
        labels = np.array([0] * 5 + [1])
        if request.param == "complete":
            dataset = MultimodalBaseDataset(data, labels)
        else:
            masks = dict(
                mod1=torch.Tensor([False] * 3 + [True] * 3),
                mod2=torch.Tensor([True] * 6),
                mod3=torch.Tensor([True] * 6),
                mod4=torch.Tensor([True] * 6),
            )
            dataset = IncompleteDataset(data=data, masks=masks, labels=labels)

        return dataset

    @pytest.fixture
    def custom_architectures(self):
        # Create an instance of mvae model
        config1 = BaseAEConfig(input_dim=(2,), latent_dim=5)
        config2 = BaseAEConfig(input_dim=(3,), latent_dim=5)
        config3 = BaseAEConfig(input_dim=(4,), latent_dim=5)

        encoders = dict(
            mod1=Encoder_VAE_MLP(config1),
            mod2=Encoder_VAE_MLP(config2),
            mod3=Encoder_VAE_MLP(config3),
            mod4=Encoder_VAE_MLP(config3),
        )

        decoders = dict(
            mod1=Decoder_AE_MLP(config1),
            mod2=Decoder_AE_MLP(config2),
            mod3=Decoder_AE_MLP(config3),
            mod4=Decoder_AE_MLP(config3),
        )

        return dict(
            encoders=encoders,
            decoders=decoders,
        )

    @pytest.fixture(params=[1.0, 1.5, 2.0])
    def model_config(self, request):
        model_config = MVTCAEConfig(
            n_modalities=4,
            latent_dim=5,
            input_dims=dict(mod1=(2,), mod2=(3,), mod3=(4,), mod4=(4,)),
            beta=request.param,
        )

        return model_config

    @pytest.fixture(params=[True, False])
    def model(self, custom_architectures, model_config, request):
        custom = request.param
        if custom:
            model = MVTCAE(model_config, **custom_architectures)
        else:
            model = MVTCAE(model_config)
        return model

    def test(self, model, dataset, model_config):
        ### Check that the grad with regard to missing modalities is null
        output = model(dataset[:3], epoch=2)
        loss = output.loss
        loss.backward()
        for param in model.encoders["mod1"].parameters():
            assert torch.all(param.grad == 0)

        output = model(dataset[-3:], epoch=2)
        loss = output.loss
        loss.backward()
        for param in model.encoders["mod1"].parameters():
            assert not torch.all(param.grad == 0)


@pytest.mark.slow
class TestTraining:
    @pytest.fixture(params=["complete", "incomplete"])
    def dataset(self, request):
        # Create simple small dataset
        data = dict(
            mod1=torch.Tensor([[1.0, 2.0], [4.0, 5.0]]),
            mod2=torch.Tensor([[67.1, 2.3, 3.0], [1.3, 2.0, 3.0]]),
            mod3=torch.Tensor([[37, 2, 4, 1], [8, 9, 7, 0]]),
            mod4=torch.Tensor([[37, 2, 4, 1], [8, 9, 7, 0]]),
        )
        labels = np.array([0, 1, 0, 0])
        if request.param == "complete":
            dataset = MultimodalBaseDataset(data, labels)
        else:
            masks = dict(
                mod1=torch.Tensor([True, False]),
                mod2=torch.Tensor([True, False]),
                mod3=torch.Tensor([True, True]),
                mod4=torch.Tensor([True, True]),
            )
            dataset = IncompleteDataset(data=data, masks=masks, labels=labels)

        return dataset

    @pytest.fixture
    def custom_architectures(self):
        # Create an instance of mmvae model
        config1 = BaseAEConfig(input_dim=(2,), latent_dim=5)
        config2 = BaseAEConfig(input_dim=(3,), latent_dim=5)
        config3 = BaseAEConfig(input_dim=(4,), latent_dim=5)

        encoders = dict(
            mod1=Encoder_VAE_MLP(config1),
            mod2=Encoder_VAE_MLP(config2),
            mod3=Encoder_VAE_MLP(config3),
            mod4=Encoder_VAE_MLP(config3),
        )

        decoders = dict(
            mod1=Decoder_AE_MLP(config1),
            mod2=Decoder_AE_MLP(config2),
            mod3=Decoder_AE_MLP(config3),
            mod4=Decoder_AE_MLP(config3),
        )

        return dict(
            encoders=encoders,
            decoders=decoders,
        )

    @pytest.fixture(params=[0.5, 1.0, 2.0])
    def model_config(self, request):
        model_config = MVTCAEConfig(
            n_modalities=4,
            latent_dim=5,
            input_dims=dict(mod1=(2,), mod2=(3,), mod3=(4,), mod4=(4,)),
            beta=request.param,
        )

        return model_config

    @pytest.fixture(params=[True, False])
    def model(self, custom_architectures, model_config, request):
        custom = request.param
        if custom:
            model = MVTCAE(model_config, **custom_architectures)
        else:
            model = MVTCAE(model_config)
        return model

    @pytest.fixture
    def training_config(self, tmpdir):
        tmpdir.mkdir("dummy_folder")
        dir_path = os.path.join(tmpdir, "dummy_folder")
        return BaseTrainerConfig(
            num_epochs=3,
            steps_saving=2,
            learning_rate=1e-4,
            optimizer_cls="AdamW",
            optimizer_params={"betas": (0.91, 0.995)},
            output_dir=dir_path,
        )

    @pytest.fixture
    def trainer(self, model, training_config, dataset):
        trainer = BaseTrainer(
            model=model,
            train_dataset=dataset,
            eval_dataset=dataset,
            training_config=training_config,
        )

        return trainer

    def new_trainer(self, model, training_config, dataset, checkpoint_dir):
        trainer = BaseTrainer(
            model=model,
            train_dataset=dataset,
            eval_dataset=dataset,
            training_config=training_config,
            checkpoint=checkpoint_dir,
        )

        return trainer

    def test_train_step(self, trainer):
        start_model_state_dict = deepcopy(trainer.model.state_dict())
        start_optimizer = trainer.optimizer
        _ = trainer.train_step(epoch=1)

        step_1_model_state_dict = deepcopy(trainer.model.state_dict())

        # check that weights were updated
        assert not all(
            [
                torch.equal(start_model_state_dict[key], step_1_model_state_dict[key])
                for key in start_model_state_dict.keys()
            ]
        )
        assert trainer.optimizer == start_optimizer

    def test_eval_step(self, trainer):
        start_model_state_dict = deepcopy(trainer.model.state_dict())

        _ = trainer.eval_step(epoch=1)

        step_1_model_state_dict = deepcopy(trainer.model.state_dict())

        # check that weights were not updated
        assert all(
            [
                torch.equal(start_model_state_dict[key], step_1_model_state_dict[key])
                for key in start_model_state_dict.keys()
            ]
        )

    def test_main_train_loop(self, trainer):
        start_model_state_dict = deepcopy(trainer.model.state_dict())

        trainer.train()

        step_1_model_state_dict = deepcopy(trainer.model.state_dict())

        # check that weights were updated
        assert not all(
            [
                torch.equal(start_model_state_dict[key], step_1_model_state_dict[key])
                for key in start_model_state_dict.keys()
            ]
        )

    def test_checkpoint_saving(self, model, trainer, training_config):
        dir_path = training_config.output_dir

        # Make a training step
        step_1_loss = trainer.train_step(epoch=1)

        model = deepcopy(trainer.model)
        optimizer = deepcopy(trainer.optimizer)

        trainer.save_checkpoint(dir_path=dir_path, epoch=0, model=model)

        checkpoint_dir = os.path.join(dir_path, "checkpoint_epoch_0")

        assert os.path.isdir(checkpoint_dir)

        files_list = os.listdir(checkpoint_dir)

        assert set(
            ["model.pt", "optimizer.pt", "training_config.json", "info_checkpoint.json"]
        ).issubset(set(files_list))

        # check pickled custom architectures
        for archi in model.model_config.custom_architectures:
            assert archi + ".pkl" in files_list

        model_rec_state_dict = torch.load(os.path.join(checkpoint_dir, "model.pt"))[
            "model_state_dict"
        ]

        assert all(
            [
                torch.equal(
                    model_rec_state_dict[key].cpu(), model.state_dict()[key].cpu()
                )
                for key in model.state_dict().keys()
            ]
        )

        # check reload full model
        model_rec = AutoModel.load_from_folder(os.path.join(checkpoint_dir))

        assert all(
            [
                torch.equal(
                    model_rec.state_dict()[key].cpu(), model.state_dict()[key].cpu()
                )
                for key in model.state_dict().keys()
            ]
        )

        assert type(model_rec.encoders.cpu()) == type(model.encoders.cpu())
        assert type(model_rec.decoders.cpu()) == type(model.decoders.cpu())

        optim_rec_state_dict = torch.load(os.path.join(checkpoint_dir, "optimizer.pt"))

        assert all(
            [
                dict_rec == dict_optimizer
                for (dict_rec, dict_optimizer) in zip(
                    optim_rec_state_dict["param_groups"],
                    optimizer.state_dict()["param_groups"],
                )
            ]
        )

        assert all(
            [
                dict_rec == dict_optimizer
                for (dict_rec, dict_optimizer) in zip(
                    optim_rec_state_dict["state"], optimizer.state_dict()["state"]
                )
            ]
        )

    def test_checkpoint_saving_during_training(
        self, model, trainer, training_config, dataset
    ):
        #
        target_saving_epoch = training_config.steps_saving

        dir_path = training_config.output_dir

        trainer.train()

        training_dir = os.path.join(
            dir_path, f"MVTCAE_training_{trainer._training_signature}"
        )

        checkpoint_dir = os.path.join(
            training_dir, f"checkpoint_epoch_{target_saving_epoch}"
        )

        # try resuming
        new_trainer_ = self.new_trainer(model, training_config, dataset, checkpoint_dir)

        assert new_trainer_.best_train_loss == trainer.best_train_loss
        assert new_trainer_.trained_epochs == target_saving_epoch

        new_trainer_.train()

    def test_resume_from_checkpoint(self, model, trainer, training_config):
        #
        target_saving_epoch = training_config.steps_saving

        dir_path = training_config.output_dir

        model = deepcopy(trainer.model)

        trainer.train()

        training_dir = os.path.join(
            dir_path, f"MVTCAE_training_{trainer._training_signature}"
        )
        assert os.path.isdir(training_dir)

        checkpoint_dir = os.path.join(
            training_dir, f"checkpoint_epoch_{target_saving_epoch}"
        )

        assert os.path.isdir(checkpoint_dir)

        files_list = os.listdir(checkpoint_dir)

        # check files
        assert set(["model.pt", "optimizer.pt", "training_config.json"]).issubset(
            set(files_list)
        )

        # check pickled custom architectures
        for archi in model.model_config.custom_architectures:
            assert archi + ".pkl" in files_list

        model_rec_state_dict = torch.load(os.path.join(checkpoint_dir, "model.pt"))[
            "model_state_dict"
        ]

        assert not all(
            [
                torch.equal(model_rec_state_dict[key], model.state_dict()[key])
                for key in model.state_dict().keys()
            ]
        )

    def test_final_model_saving(self, model, trainer, training_config):
        dir_path = training_config.output_dir

        trainer.train()

        model = deepcopy(trainer._best_model)

        training_dir = os.path.join(
            dir_path, f"MVTCAE_training_{trainer._training_signature}"
        )
        assert os.path.isdir(training_dir)

        final_dir = os.path.join(training_dir, f"final_model")
        assert os.path.isdir(final_dir)

        files_list = os.listdir(final_dir)

        assert set(["model.pt", "model_config.json", "training_config.json"]).issubset(
            set(files_list)
        )

        # check pickled custom architectures
        for archi in model.model_config.custom_architectures:
            assert archi + ".pkl" in files_list

        # check reload full model
        model_rec = AutoModel.load_from_folder(os.path.join(final_dir))

        assert all(
            [
                torch.equal(
                    model_rec.state_dict()[key].cpu(), model.state_dict()[key].cpu()
                )
                for key in model.state_dict().keys()
            ]
        )

        assert type(model_rec.encoders.cpu()) == type(model.encoders.cpu())
        assert type(model_rec.decoders.cpu()) == type(model.decoders.cpu())

    def test_compute_nll(self, model, dataset):
        if not hasattr(dataset, "masks"):
            nll = model.compute_joint_nll(dataset, K=10, batch_size_K=2)
            assert nll >= 0
            assert type(nll) == torch.Tensor
            assert nll.size() == torch.Size([])

            cnll = model.cond_nll_from_subset(
                dataset, ["mod1", "mod2"], ["mod3"], K=10, batch_size_k=2
            )
            assert type(cnll) == dict
            assert "mod3" in cnll.keys()



import os
import shutil

import torch
from pythae.data.datasets import BaseDataset
from pythae.models.normalizing_flows import IAF, IAFConfig, NFModel
from pythae.trainers import BaseTrainer, BaseTrainerConfig
from torch.distributions import MultivariateNormal
from torch.nn import ModuleDict
from torch.utils.data import DataLoader

from multivae.data.utils import set_inputs_to_device
from multivae.models.base import ModelOutput

from ...models import BaseMultiVAE
from ..base.base_sampler import BaseSampler
from .iaf_sampler_config import IAFSamplerConfig


class IAFSampler(BaseSampler):
    """Fits an Inverse Autoregressive Flow in the multimodal autoencoder's latent space.
    If the model has multiple latent spaces, we fit one flow per latent space.

    Args:
        model (BaseMultiVAE): The model to sample from
        sampler_config (IAFSamplerConfig): A IAFSamplerConfig instance containing
            the main parameters of the sampler. If None, a pre-defined configuration is used.
            Default: None

    .. note::

        The method :class:`~multivae.samplers.IAFSampler.fit` must be called to fit the
        sampler before sampling.
    """

    def __init__(self, model: BaseMultiVAE, sampler_config: IAFSamplerConfig = None):
        self.is_fitted = False

        if sampler_config is None:
            sampler_config = IAFSamplerConfig()

        BaseSampler.__init__(self, model=model, sampler_config=sampler_config)

        self.flows_dims = dict(shared=model.model_config.latent_dim)
        if self.model.multiple_latent_spaces:
            self.flows_dims.update(self.model.style_dims)

        self.priors = dict()
        self.flows_models = dict()

        for key in self.flows_dims:
            self.priors[key] = MultivariateNormal(
                torch.zeros(self.flows_dims[key]).to(self.device),
                torch.eye(self.flows_dims[key]).to(self.device),
            )

            iaf_config = IAFConfig(
                input_dim=(self.flows_dims[key],),
                n_made_blocks=sampler_config.n_made_blocks,
                n_hidden_in_made=sampler_config.n_hidden_in_made,
                hidden_size=sampler_config.hidden_size,
                include_batch_norm=sampler_config.include_batch_norm,
            )

            iaf_model = IAF(model_config=iaf_config)
            self.flows_models[key] = NFModel(self.priors[key], iaf_model).to(
                self.device
            )

        self.name = "IAFsampler"

    def fit(
        self, train_data, eval_data=None, training_config: BaseTrainerConfig = None
    ):
        """Method to fit the sampler from the training data.

        Args:
            train_data (MultimodalBaseDataset): The train data needed to retreive the training embeddings
                    and fit the mixture in the latent space. Must be of shape n_imgs x im_channels x
                    ... and in range [0-1]
            eval_data (MultimodalBaseDataset): The train data needed to retreive the evaluation embeddings
                    and fit the mixture in the latent space. Must be of shape n_imgs x im_channels x
                    ... and in range [0-1]
            training_config (BaseTrainerConfig): the training config to use to fit the flow.
        """

        train_loader = DataLoader(dataset=train_data, batch_size=100, shuffle=True)

        zs = {m: [] for m in self.flows_models}

        try:
            with torch.no_grad():
                for _, inputs in enumerate(train_loader):
                    inputs = set_inputs_to_device(inputs, self.device)
                    encoder_output = self.model.encode(inputs)
                    zs["shared"].append(encoder_output.z)

                    if self.model.multiple_latent_spaces:
                        for m in encoder_output.modalities_z:
                            zs[m].append(encoder_output.modalities_z[m])

        except RuntimeError:
            for _, inputs in enumerate(train_loader):
                inputs = set_inputs_to_device(inputs, self.device)
                encoder_output = self.model.encode(inputs)
                zs["shared"].append(encoder_output.z.detach())

                if self.model.multiple_latent_spaces:
                    for m in encoder_output.modalities_z:
                        zs[m].append(encoder_output.modalities_z[m].detach())

        train_data = {m: torch.cat(zs[m], dim=0) for m in zs}

        # Eval dataset

        if eval_data is not None:
            eval_loader = DataLoader(dataset=eval_data, batch_size=100, shuffle=False)

            zs = {m: [] for m in self.flows_models}

            try:
                with torch.no_grad():
                    for _, inputs in enumerate(eval_loader):
                        inputs = set_inputs_to_device(inputs, self.device)
                        encoder_output = self.model.encode(inputs)
                        zs["shared"].append(encoder_output.z)
                        if self.model.multiple_latent_spaces:
                            for m in encoder_output.modalities_z:
                                zs[m].append(encoder_output.modalities_z[m])

            except RuntimeError:
                for _, inputs in enumerate(train_loader):
                    inputs = set_inputs_to_device(inputs, self.device)
                    encoder_output = self.model.encode(inputs)
                    zs["shared"].append(encoder_output.z.detach())
                    if self.model.multiple_latent_spaces:
                        for m in encoder_output.modalities_z:
                            zs[m].append(encoder_output.modalities_z[m].detach())

            eval_data = {m: torch.cat(zs[m]) for m in zs}

        self.iaf_models = torch.nn.ModuleDict()

        for m in train_data:  # number of latent_spaces
            train_dataset = BaseDataset(
                data=train_data[m], labels=torch.zeros((len(train_data[m]),))
            )
            eval_dataset = (
                None
                if eval_data is None
                else BaseDataset(
                    data=eval_data[m], labels=torch.zeros((len(eval_data[m]),))
                )
            )

            trainer = BaseTrainer(
                model=self.flows_models[m],
                train_dataset=train_dataset,
                eval_dataset=eval_dataset,
                training_config=training_config,
            )

            trainer.train()

            self.iaf_models[m] = IAF.load_from_folder(
                os.path.join(trainer.training_dir, "final_model")
            ).to(self.device)

            shutil.rmtree(trainer.training_dir)

        self.is_fitted = True

    def sample(
        self, n_samples: int = 1, batch_size: int = 500, **kwargs
    ) -> torch.Tensor:
        """Main sampling function of the sampler.

        Args:
            num_samples (int): The number of samples to generate
            batch_size (int): The batch size to use during sampling
            output_dir (str): The directory where the images will be saved. If does not exist the
                folder is created. If None: the images are not saved. Defaults: None.
            return_gen (bool): Whether the sampler should directly return a tensor of generated
                data. Default: True.
            save_sampler_config (bool): Whether to save the sampler config. It is saved in
                output_dir

        Returns:
            ~torch.Tensor: The generated images
        """

        if not self.is_fitted:
            raise ArithmeticError(
                "The sampler needs to be fitted by calling sampler.fit() method"
                "before sampling."
            )

        full_batch_nbr = int(n_samples / batch_size)
        last_batch_samples_nbr = n_samples % batch_size
        batches = [batch_size] * full_batch_nbr
        if last_batch_samples_nbr != 0:
            batches = batches + [last_batch_samples_nbr]

        z_gen = {m: [] for m in self.iaf_models}

        for batch in batches:
            for m in self.iaf_models:
                u = self.priors[m].sample((batch,))
                z = self.iaf_models[m].inverse(u).out
                z_gen[m].append(z)

        # Output with the same format as the output of encode or generate_from_prior functions
        output = ModelOutput(
            z=torch.cat(z_gen.pop("shared")),
            one_latent_space=not self.model.multiple_latent_spaces,
        )
        if self.model.multiple_latent_spaces:
            output["modalities_z"] = {m: torch.cat(z_gen[m]) for m in z_gen}

        return output



import logging
import os
from pathlib import Path
from typing import Union

import numpy as np
import torch
from torchvision.datasets import MNIST, SVHN

from .base import MultimodalBaseDataset
from .utils import ResampleDataset

logger = logging.getLogger(__name__)
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class MnistSvhn(MultimodalBaseDataset): # pragma: no cover
    """

    A paired MnistSvhn dataset.

    Args:
        path (str) : The path where the data is saved.
        split (str) : Either 'train' or 'test'.
        download (bool) : Whether to download the data or not. Default to True.
        data_multiplication (int) :

        **kwargs:

            transform_mnist (Transform) : a transformation to apply to MNIST. If none specified, a simple ToTensor() is applied.
            transform_svhn (Transform) : a transformation to apply to SVHN. If none specified, a simple ToTensor() is applied.

    """

    def __init__(
        self,
        data_path: Union[str, Path],
        split: str = "train",
        download=False,
        data_multiplication=5,
        **kwargs,
    ):
        if split not in ["train", "test"]:
            raise AttributeError("Possible values for split are 'train' or 'test'")

        # Load unimodal datasets

        mnist = MNIST(data_path, train=(split == "train"), download=download)
        svhn = SVHN(data_path, split=split, download=download)

        self.data_mul = data_multiplication
        self.path_to_idx = (
            data_path + f"/mnist_svhn_idx_data_mul_{self.data_mul}/" + split
        )
        # Check if a pairing already exists and if not create one
        if not self._check_pairing_exists():
            self.create_pairing(mnist, svhn)

        i_mnist = torch.load(f"{self.path_to_idx}/mnist_idx.pt")  ## !!!!WARNING!!!
        i_svhn = torch.load(f"{self.path_to_idx}/svhn_idx.pt")  ## !!!!WARNING!!!

        order = np.arange(len(i_mnist))
        np.random.shuffle(
            order
        )  # shuffle the samples so that they are not ordered by labels.
        labels = mnist.targets[i_mnist][order]

        # Resample the datasets

        data_mnist = mnist.data.float().div(255).unsqueeze(1)
        data_svhn = torch.FloatTensor(svhn.data).div(255)
        mnist = ResampleDataset(
            data_mnist, lambda d, i: i_mnist[order[i]], size=len(i_mnist)
        )
        svhn = ResampleDataset(
            data_svhn, lambda d, i: i_svhn[order[i]], size=len(i_svhn)
        )
        data = dict(mnist=mnist, svhn=svhn)

        self.data_path = data_path
        super().__init__(data, labels)

    def _check_pairing_exists(self):
        if not os.path.exists(f"{self.path_to_idx}/mnist_idx.pt"):
            logger.warning("Pairing not found.")
            return False
        if not os.path.exists(f"{self.path_to_idx}/svhn_idx.pt"):
            logger.warning("Pairing not found.")
            return False
        return True

    def rand_match_on_idx(self, l1, idx1, l2, idx2, max_d=10000):
        _idx1, _idx2 = [], []
        for l in l1.unique():  # assuming both have same idxs
            l_idx1, l_idx2 = idx1[l1 == l], idx2[l2 == l]
            n = min(l_idx1.size(0), l_idx2.size(0), max_d)
            l_idx1, l_idx2 = l_idx1[:n], l_idx2[:n]
            for _ in range(self.data_mul):
                _idx1.append(l_idx1[torch.randperm(n)])
                _idx2.append(l_idx2[torch.randperm(n)])
        return torch.cat(_idx1), torch.cat(_idx2)

    def create_pairing(self, mnist: MNIST, svhn: SVHN, max_d=10000):
        logger.info(f"Creating indices in {self.path_to_idx}")
        # Refactor svhn labels to match mnist labels
        svhn.labels = torch.LongTensor(svhn.labels.squeeze().astype(int)) % 10
        mnist_l, mnist_li = mnist.targets.sort()
        svhn_l, svhn_li = svhn.labels.sort()
        idx1, idx2 = self.rand_match_on_idx(
            mnist_l, mnist_li, svhn_l, svhn_li, max_d=max_d
        )

        path = Path(self.path_to_idx)
        path.mkdir(parents=True, exist_ok=True)
        torch.save(idx1, f"{self.path_to_idx}/mnist_idx.pt")
        torch.save(idx2, f"{self.path_to_idx}/svhn_idx.pt")



from pydantic.dataclasses import dataclass

from ..joint_models import BaseJointModelConfig


@dataclass
class JNFDccaConfig(BaseJointModelConfig):
    """
    This is the base config for the JNFDcca model.

    Args:
        n_modalities (int): The number of modalities. Default: None.
        latent_dim (int): The dimension of the latent space. Default: None.
        input_dims (dict[str,tuple]) : The modalities'names (str) and input shapes (tuple).
        uses_likelihood_rescaling (bool): To mitigate modality collapse, it is possible to use likelihood rescaling.
            (see : https://proceedings.mlr.press/v162/javaloy22a.html).
            The inputs_dim must be provided to compute the likelihoods rescalings. Default to False.
        decoders_dist (Dict[str, Union[function, str]]). The decoder distributions to use per modality.
            Per modality, you can provide a string in ['normal','bernoulli','laplace']. If None is provided,
            a normal distribution is used for each modality.
        decoder_dist_params (Dict[str,dict]) : Parameters for the output decoder distributions, for
            computing the log-probability.
            For instance, with normal or laplace distribution, you can pass the scale in this dictionary.
            ex :  {'mod1' : {scale : 0.75}}
        warmup (int): The number of warmup epochs during training. Default to 10.

        nb_epochs_dcca (int) : The number of epochs during which to train the DCCA embeddings. Default to 30.
        embedding_dcca_dim (int) : The dimension of the DCCA embedding to use. Default to 20.
        use_all_singular_values (bool) : Wether to use all the singular values for the computation of the objective.
            Using True is more unstable. Default to False.

    """

    warmup: int = 10
    nb_epochs_dcca: int = 30
    embedding_dcca_dim: int = 20
    use_all_singular_values: bool = (
        False  # Using True generally leads to NaN in the loss.
    )



import importlib

model_card_template = """---
language: en
tags:
- multivae
license: apache-2.0
---

### Downloading this model from the Hub
This model was trained with multivae. It can be downloaded or reloaded using the method `load_from_hf_hub`
```python
>>> from multivae.models import AutoModel
>>> model = AutoModel.load_from_hf_hub(hf_hub_path="your_hf_username/repo_name")
```
"""


def hf_hub_is_available():
    return importlib.util.find_spec("huggingface_hub") is not None



import pytest
import torch
from pythae.models.base.base_utils import ModelOutput

from multivae.data.datasets import MultimodalBaseDataset
from multivae.metrics import FIDEvaluator, FIDEvaluatorConfig
from multivae.metrics.fids.fids import adapt_shape_for_fid
from multivae.models import MVTCAE, MVTCAEConfig
from multivae.models.nn.default_architectures import Encoder_VAE_MLP
from multivae.samplers import GaussianMixtureSampler, GaussianMixtureSamplerConfig


# @pytest.mark.slow
class TestFIDMetrics:
    @pytest.fixture
    def model(self):
        model_config = MVTCAEConfig(
            n_modalities=2,
            input_dims={"m0": (3, 32, 32), "m1": (1, 28, 28)},
        )
        return MVTCAE(model_config)

    @pytest.fixture
    def dataset(self):
        return MultimodalBaseDataset(
            data={
                "m0": torch.randn((128, 3, 32, 32)),
                "m1": torch.randn((128, 1, 28, 28)),
            },
            labels=torch.ones((1024,)),
        )

    @pytest.fixture(params=["custom_config", "default_config"])
    def config(self, request):
        config = FIDEvaluatorConfig(
            batch_size=64, resize=False, inception_weights_path="./fids.pt"
        )
        if request.param == "custom_config":
            return config
        else:
            return FIDEvaluatorConfig(batch_size=64)

    @pytest.fixture(params=[True, False])
    def sampler(self, request, model, dataset):
        if not request.param:
            return None
        else:
            sampler_config = GaussianMixtureSamplerConfig()
            sampler = GaussianMixtureSampler(model, sampler_config)
            sampler.fit(dataset)
            return sampler

    @pytest.fixture(params=[None, lambda x: adapt_shape_for_fid(resize=False)(x)])
    def fid_model(self, model, dataset, config, sampler, request):
        return FIDEvaluator(
            model=model,
            test_dataset=dataset,
            output=None,
            eval_config=config,
            custom_encoder=None,
            transform=request.param,
            sampler=sampler,
        )  # Add test with custom encoder

    def test(self, fid_model, config):
        assert (fid_model.n_data) == 128
        assert len(fid_model.test_loader) > 0
        assert fid_model.batch_size == config.batch_size
        output = fid_model.eval()
        assert isinstance(output, ModelOutput)



"""
**Abstract class**
"""

from pythae.models.base.base_utils import ModelOutput
from pythae.models.nn.base_architectures import BaseDecoder, BaseEncoder

from .base_config import BaseAEConfig, BaseMultiVAEConfig
from .base_model import BaseMultiVAE

__all__ = [
    "BaseMultiVAEConfig",
    "BaseAEConfig",
    "BaseMultiVAE",
    "BaseEncoder",
    "BaseDecoder",
    "ModelOutput",
]



import logging
from typing import Union

import numpy as np
import torch
import torch.distributions as dist
from pythae.models.base.base_utils import ModelOutput
from pythae.models.nn.base_architectures import BaseEncoder

from ...data.datasets.base import MultimodalBaseDataset
from ..joint_models import BaseJointModel
from .jmvae_config import JMVAEConfig

logger = logging.getLogger(__name__)
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class JMVAE(BaseJointModel):

    """The Joint Multimodal Variational Autoencoder model.

    Args:
        model_config (JMVAEConfig): An instance of JMVAEConfig in which any model's
            parameters is made available.

        encoders (Dict[str, ~pythae.models.nn.base_architectures.BaseEncoder]): A dictionary containing
            the modalities names and the encoders for each modality. Each encoder is an instance of
            Pythae's BaseEncoder. Default: None.

        decoder (Dict[str, ~pythae.models.nn.base_architectures.BaseDecoder]): A dictionary containing
            the modalities names and the decoders for each modality. Each decoder is an instance of
            Pythae's BaseDecoder.

        joint_encoder (~pythae.models.nn.base_architectures.BaseEncoder) : An instance of
            BaseEncoder that takes all the modalities as an input. If none is provided, one is
            created from the unimodal encoders. Default : None.
    """

    def __init__(
        self,
        model_config: JMVAEConfig,
        encoders: dict = None,
        decoders: dict = None,
        joint_encoder: Union[BaseEncoder, None] = None,
        **kwargs,
    ):
        super().__init__(model_config, encoders, decoders, joint_encoder, **kwargs)

        self.model_name = "JMVAE"

        self.alpha = model_config.alpha
        self.warmup = model_config.warmup

    def encode(
        self,
        inputs: MultimodalBaseDataset,
        cond_mod: Union[list, str] = "all",
        N: int = 1,
        **kwargs,
    ):
        """Compute encodings from the posterior distributions conditioning on all the modalities or
        a subset of the modalities.

        Args:
            inputs (MultimodalBaseDataset): The data to encode.
            cond_mod (Union[list, str], optional): The modalities to use to compute the posterior
            distribution. Defaults to 'all'.
            N (int, optional): The number of samples to generate from the posterior distribution
            for each datapoint. Defaults to 1.

        Raises:
            AttributeError: _description_
            AttributeError: _description_

        Returns:
            ModelOutput: Containing z the latent variables.
        """
        self.eval()

        cond_mod = super().encode(inputs, cond_mod, N, **kwargs).cond_mod
        sample_shape = [] if N == 1 else [N]
        return_mean = kwargs.pop("return_mean", False)

        if len(cond_mod) == self.n_modalities:
            output = self.joint_encoder(inputs.data)
            if return_mean:
                z = torch.stack([output.embedding] * N) if N > 1 else output.embedding
            else:
                z = dist.Normal(
                    output.embedding, torch.exp(0.5 * output.log_covariance)
                ).rsample(sample_shape)
            if N > 1 and kwargs.pop("flatten", False):
                N, l, d = z.shape
                z = z.reshape(l * N, d)
            return ModelOutput(z=z, one_latent_space=True)

        elif len(cond_mod) != 1:
            z = self.sample_from_poe_subset_exact(
                cond_mod, inputs.data, sample_shape=sample_shape
            )

            if N > 1 and kwargs.pop("flatten", False):
                N, l, d = z.shape
                z = z.reshape(l * N, d)
            return ModelOutput(z=z, one_latent_space=True)

        elif len(cond_mod) == 1:
            cond_mod = cond_mod[0]
            output = self.encoders[cond_mod](inputs.data[cond_mod])
            sample_shape = [] if N == 1 else [N]

            z = dist.Normal(
                output.embedding, torch.exp(0.5 * output.log_covariance)
            ).rsample(
                sample_shape
            )  # shape N x len_data x latent_dim
            if N > 1 and kwargs.pop("flatten", False):
                N, l, d = z.shape
                z = z.reshape(l * N, d)
            return ModelOutput(z=z, one_latent_space=True)
        else:
            raise AttributeError(
                f"Modalities of names {cond_mod} not handled. The"
                f" modalities that can be encoded are {list(self.encoders.keys())}"
            )

    def forward(self, inputs: MultimodalBaseDataset, **kwargs) -> ModelOutput:
        """Performs a forward pass of the JMVAE model on inputs.

        Args:
            inputs (MultimodalBaseDataset)
            warmup (int) : number of warmup epochs to do. The weigth of the regularization augments
                linearly to reach 1 at the end of the warmup. The enforces the optimization of
                the reconstruction term only at first.
            epoch (int) : the epoch number during which forward is called.

        Returns:
            ModelOutput
        """

        epoch = kwargs.pop("epoch", 1)

        # Compute the reconstruction term
        joint_output = self.joint_encoder(inputs.data)
        mu, log_var = joint_output.embedding, joint_output.log_covariance

        sigma = torch.exp(0.5 * log_var)
        qz_xy = dist.Normal(mu, sigma)
        z_joint = qz_xy.rsample()

        recon_loss = 0

        # Decode in each modality
        len_batch = 0
        for mod in self.decoders:
            x_mod = inputs.data[mod]
            len_batch = len(x_mod)
            recon_mod = self.decoders[mod](z_joint).reconstruction
            recon_loss += (
                -self.recon_log_probs[mod](recon_mod, x_mod) * self.rescale_factors[mod]
            ).sum()

        # Compute the KLD to the prior
        KLD = -0.5 * torch.sum(1 + log_var - mu.pow(2) - log_var.exp())

        # Compute the KL between unimodal and joint encoders
        LJM = 0
        for mod in self.encoders:
            output = self.encoders[mod](inputs.data[mod])
            uni_mu, uni_log_var = output.embedding, output.log_covariance
            LJM += (
                1
                / 2
                * (
                    uni_log_var
                    - log_var
                    + (torch.exp(log_var) + (mu - uni_mu) ** 2) / torch.exp(uni_log_var)
                    - 1
                )
            )

        LJM = LJM.sum() * self.alpha

        # Compute the total loss to minimize

        reg_loss = KLD + LJM
        if epoch >= self.warmup:
            beta = 1
        else:
            beta = epoch / self.warmup
        recon_loss, reg_loss = recon_loss / len_batch, reg_loss / len_batch
        elbo = (recon_loss + KLD) / len_batch
        loss = recon_loss + beta * reg_loss

        metrics = dict(loss_no_ponderation=reg_loss + recon_loss, beta=beta, elbo=elbo)

        output = ModelOutput(loss=loss, metrics=metrics)

        return output

    def poe(self, mus_list, log_vars_list):
        mus = mus_list.copy()
        log_vars = log_vars_list.copy()

        # Add the prior to the product of experts
        mus.append(torch.zeros_like(mus[0]))
        log_vars.append(torch.zeros_like(log_vars[0]))

        # Compute the joint posterior
        lnT = torch.stack([-l for l in log_vars])  # Compute the inverse of variances
        lnV = -torch.logsumexp(lnT, dim=0)  # variances of the product of expert
        mus = torch.stack(mus)
        joint_mu = (torch.exp(lnT) * mus).sum(dim=0) * torch.exp(lnV)

        return joint_mu, lnV

    def sample_from_poe_subset_exact(self, subset: list, data: dict, sample_shape=[]):
        """
        Sample from the product of experts for infering from a subset of modalities.
        """

        mus, logvars = [], []

        for mod in subset:
            vae_output = self.encoders[mod](data[mod])
            mu, log_var = vae_output.embedding, vae_output.log_covariance
            mus.append(vae_output.embedding)
            logvars.append(vae_output.log_covariance)

        joint_mu, joint_logvar = self.poe(mus, logvars)
        z = dist.Normal(joint_mu, torch.exp(0.5 * joint_logvar)).rsample(sample_shape)
        return z

    def compute_joint_nll_paper(
        self, inputs: MultimodalBaseDataset, K: int = 1000, batch_size_K: int = 100
    ):
        """
        Return the estimated negative log-likelihood summed over the input batch.
        The negative log-likelihood is estimated using importance sampling.

        Args:
            inputs : the data to compute the joint likelihood"""

        # First compute all the parameters of the joint posterior q(z|x,y)

        joint_output = self.joint_encoder(inputs.data)
        mu, log_var = joint_output.embedding, joint_output.log_covariance

        sigma = torch.exp(0.5 * log_var)
        qz_xy = dist.Normal(mu, sigma)
        # And sample from the posterior
        z_joint = qz_xy.rsample([K])  # shape K x n_data x latent_dim
        z_joint = z_joint.permute(1, 0, 2)
        n_data, _, latent_dim = z_joint.shape

        # Then iter on each datapoint to compute the iwae estimate of ln(p(x))
        ll = 0

        for i in range(n_data):
            start_idx = 0
            stop_idx = min(start_idx + batch_size_K, K)
            lnpxs = []
            while start_idx < stop_idx:
                latents = z_joint[i][start_idx:stop_idx]

                # Compute p(x_m|z) for z in latents and for each modality m
                lpx_zs = 0  # ln(p(x,y|z))
                for mod in ["images"]:  # only keep the images for this likelihood
                    decoder = self.decoders[mod]
                    recon = decoder(latents)[
                        "reconstruction"
                    ]  # (batch_size_K, *decoder_output_shape)
                    x_m = inputs.data[mod][i]  # (*input_shape)
                    lpx_zs += (
                        self.recon_log_probs[mod](recon, x_m)
                        .reshape(recon.size(0), -1)
                        .sum(-1)
                    )

                # Compute ln(p(z))
                prior = dist.Normal(0, 1)
                lpz = prior.log_prob(latents).sum(dim=-1)

                # Compute posteriors -ln(q(z|x,y))
                qz_xy = dist.Normal(mu[i], sigma[i])
                lqz_xy = qz_xy.log_prob(latents).sum(dim=-1)

                ln_px = torch.logsumexp(lpx_zs + lpz - lqz_xy, dim=0)
                lnpxs.append(ln_px)

                # next batch
                start_idx += batch_size_K
                stop_idx = min(start_idx + batch_size_K, K)

            ll += torch.logsumexp(torch.Tensor(lnpxs), dim=0) - np.log(K)

        return -ll



"""
Multimodal dataset wrapper for the MNIST labels dataset.
"""
from typing import Literal

import torch
from torch.distributions import Bernoulli
from torchvision.datasets import MNIST

from .base import DatasetOutput, MultimodalBaseDataset


class MnistLabels(MultimodalBaseDataset): # pragma: no cover
    def __init__(
        self, data_path: str, split: Literal["train", "test"] = "train",
        download=False,
        random_binarized=True,
        dtype = torch.float32
    ):
        """
        Class to wrap the MnistLabels dataset for MultiVae use. 

        Args:
            data_path (str): Where is stored the data. 
            split (Literal["train", "test"], optional): The split to use. Defaults to "train".
            download (bool, optional): If the data is not available at the given data_path, whether to download it or not.
                Defaults to False.
            random_binarized (bool, optional): Whether binarize the images using a Bernoulli distribution for each pixel.
                Defaults to True.
            dtype (_type_, optional): Type for the arrays. Defaults to torch.float32.
        """
        torchvision_dataset = MNIST(
            root=data_path, train=(split == "train"), download=download
        )

        self.images = torchvision_dataset.data.div(255).unsqueeze(1).to(dtype)
        self.labels = torchvision_dataset.targets
        self.labels_one_hot =  torch.zeros(len(self.labels), 10)
        self.labels_one_hot = self.labels_one_hot.scatter(1, self.labels.unsqueeze(1), 1)
        self.labels_one_hot = self.labels_one_hot.unsqueeze(1)
        self.random_binarized = random_binarized

    def __getitem__(self, index):
        if self.random_binarized:
            images = Bernoulli(self.images[index]).sample()
        else :
            images = self.images[index]
        return DatasetOutput(
            
            data=dict(
                images=images, labels=self.labels_one_hot[index]
            ),
            labels=self.labels[index],
        )

    def __len__(self):
        return len(self.labels)



import os

import torch
from PIL import Image
from torch.utils.data import DataLoader
from torchvision.utils import make_grid

from multivae.data import MultimodalBaseDataset
from multivae.data.datasets.utils import adapt_shape
from multivae.data.utils import set_inputs_to_device
from multivae.metrics.base.evaluator_config import EvaluatorConfig
from multivae.models.base import BaseMultiVAE, ModelOutput
from multivae.samplers.base import BaseSampler

from ..base.evaluator_class import Evaluator
from .visualize_config import VisualizationConfig


class Visualization(Evaluator):

    """
    Visualization Module for visualizing unconditional, conditional samples from models.

    Args:

        model (BaseMultiVAE) : the model to evaluate.
        test_dataset (MultimodalBaseDataset) : the dataset to use for conditional image generation.
        output (str): the path where to save images and metrics. Default to None.
        eval_config (VisualizationConfig) : The configuration file for this evaluation module.
            Optional.
        sampler (BaseSampler) : The sampler to use for joint generation. Optional. If None is
            provided, the sampler is used.
    """

    def __init__(
        self,
        model: BaseMultiVAE,
        test_dataset: MultimodalBaseDataset,
        output: str = None,
        eval_config=VisualizationConfig(),
        sampler: BaseSampler = None,
    ) -> None:
        super().__init__(model, test_dataset, output, eval_config, sampler)
        self.n_samples = eval_config.n_samples
        self.n_data_cond = eval_config.n_data_cond

    def unconditional_samples(self, **kwargs):
        device = kwargs.pop("device", "cuda" if torch.cuda.is_available() else "cpu")
        if self.sampler is None:
            samples = self.model.generate_from_prior(self.n_samples)
        else:
            samples = self.sampler.sample(self.n_samples)
        from multivae.data.utils import set_inputs_to_device

        samples = set_inputs_to_device(samples, device=device)
        images = self.model.decode(samples)
        recon, shape = adapt_shape(images)

        recon_image = torch.cat(list(recon.values()))

        # Transform to PIL format
        recon_image = make_grid(recon_image, nrow=self.n_samples)
        # Add 0.5 after unnormalizing to [0, 255] to round to nearest integer
        ndarr = (
            recon_image.mul(255)
            .add_(0.5)
            .clamp_(0, 255)
            .permute(1, 2, 0)
            .to("cpu", torch.uint8)
            .numpy()
        )
        recon_image = Image.fromarray(ndarr)

        if self.output is not None:
            recon_image.save(os.path.join(self.output, "unconditional.png"))

        if self.wandb_run is not None:
            import wandb

            self.wandb_run.log({"unconditional_generation": wandb.Image(recon_image)})

        return recon_image

    def conditional_samples_subset(self, subset):
        dataloader = DataLoader(self.test_dataset, batch_size=self.n_data_cond)
        data = next(iter(dataloader))
        # set inputs to device
        data = set_inputs_to_device(data, self.device)

        recon = self.model.predict(
            data, subset, "all", N=self.n_samples, flatten=True, ignore_incomplete=True
        )
        recon.update({f"original_{m}": data.data[m] for m in subset})
        recon, shape = adapt_shape(recon)
        recon_image = [recon[f"original_{m}"] for m in subset]

        recon_image = recon_image + [recon[m] for m in recon if "original" not in m]
        recon_image = torch.cat(recon_image)

        # Transform to PIL format
        recon_image = make_grid(recon_image, nrow=self.n_data_cond)
        # Add 0.5 after unnormalizing to [0, 255] to round to nearest integer
        ndarr = (
            recon_image.mul(255)
            .add_(0.5)
            .clamp_(0, 255)
            .permute(1, 2, 0)
            .to("cpu", torch.uint8)
            .numpy()
        )
        recon_image = Image.fromarray(ndarr)

        if self.output is not None:
            recon_image.save(
                os.path.join(self.output, f"conditional_from_subset_{subset}.png")
            )

        if self.wandb_run is not None:
            import wandb

            self.wandb_run.log(
                {f"conditional_from_subset_{subset}": wandb.Image(recon_image)}
            )

        return recon_image

    def eval(self):
        image = self.unconditional_samples()

        return ModelOutput(unconditional_generation=image)



"""Test functions in the MMVAE model"""

import math

import numpy as np
import pytest
import torch
import torch.distributions as dist

from multivae.data.datasets.base import MultimodalBaseDataset
from multivae.models import MMVAE, MMVAEConfig


def is_multidata(dataB):
    return isinstance(dataB, list) or isinstance(dataB, tuple)


def log_mean_exp(value, dim=0, keepdim=False):
    return torch.logsumexp(value, dim, keepdim=keepdim) - math.log(value.size(dim))


def compute_microbatch_split(x, K):
    """Checks if batch needs to be broken down further to fit in memory."""
    B = x[0].size(0) if is_multidata(x) else x.size(0)
    S = (
        sum([1.0 / (K * math.prod(_x.size()[1:])) for _x in x])
        if is_multidata(x)
        else 1.0 / (K * math.prod(x.size()[1:]))
    )
    S = int(1e8 * S)  # float heuristic for 12Gb cuda memory
    assert S > 0, "Cannot fit individual data in memory, consider smaller K"
    return min(B, S)


def _m_dreg_looser(model, x, K=1):
    """DERG estimate for log p_\theta(x) for multi-modal vae -- fully vectorised
    This version is the looser bound---with the average over modalities outside the log
    """
    qz_xs, px_zs, zss = model(x, K)
    qz_xs_ = [vae.qz_x(*[p.detach() for p in vae.qz_x_params]) for vae in model.vaes]
    lws = []
    for r, vae in enumerate(model.vaes):
        lpz = model.pz(*model.pz_params).log_prob(zss[r]).sum(-1)
        lqz_x = log_mean_exp(
            torch.stack([qz_x_.log_prob(zss[r]).sum(-1) for qz_x_ in qz_xs_])
        )
        lpx_z = [
            px_z.log_prob(x[d])
            .view(*px_z.batch_shape[:2], -1)
            .mul(model.vaes[d].llik_scaling)
            .sum(-1)
            for d, px_z in enumerate(px_zs[r])
        ]
        lpx_z = torch.stack(lpx_z).sum(0)
        lw = lpz + lpx_z - lqz_x
        lws.append(lw)
    return torch.stack(lws), torch.stack(zss)


def _m_dreg_looser_test(qz_xs_, px_zs, zss, model: MMVAE, x, rescale_factors, K=1):
    """DERG estimate for log p_\theta(x) for multi-modal vae -- fully vectorised
    This version is the looser bound---with the average over modalities outside the log
    """
    lws = []
    for r, vae in enumerate(model.encoders):
        lpz = model.prior_dist(*model.pz_params).log_prob(zss[r]).sum(-1)
        lqz_x = log_mean_exp(
            torch.stack([qz_x_.log_prob(zss[r]).sum(-1) for qz_x_ in qz_xs_])
        )
        lpx_z = [
            px_z.log_prob(x[d])
            .view(*px_z.batch_shape[:2], -1)
            .mul(rescale_factors[d])
            .sum(-1)
            for d, px_z in enumerate(px_zs[r])
        ]
        lpx_z = torch.stack(lpx_z).sum(0)
        lw = lpz + lpx_z - lqz_x
        lws.append(lw)
    lw = torch.stack(lws)
    zss = torch.stack(zss)
    with torch.no_grad():
        grad_wt = (lw - torch.logsumexp(lw, 1, keepdim=True)).exp()
        if zss.requires_grad:
            zss.register_hook(lambda grad: grad_wt.unsqueeze(-1) * grad)
    return -(grad_wt * lw).mean(0).sum()


def _m_iwae_test(qz_xs, px_zs, zss, model: MMVAE, x, rescale_factors, K=1):
    """IWAE estimate for log p_\theta(x) for multi-modal vae -- fully vectorised
    This version is the looser bound---with the average over modalities outside the log
    """
    lws = []
    for r, qz_x in enumerate(qz_xs):
        lpz = model.prior_dist(*model.pz_params).log_prob(zss[r]).sum(-1)
        lqz_x = log_mean_exp(
            torch.stack([qz_x.log_prob(zss[r]).sum(-1) for qz_x in qz_xs])
        )
        lpx_z = [
            px_z.log_prob(x[d])
            .view(*px_z.batch_shape[:2], -1)
            .mul(rescale_factors[d])
            .sum(-1)
            for d, px_z in enumerate(px_zs[r])
        ]
        lpx_z = torch.stack(lpx_z).sum(0)
        lw = lpz + lpx_z - lqz_x

        lws.append(lw)
    lws = torch.cat(lws)  # (n_modality * n_samples) x batch_size

    return log_mean_exp(lws, dim=0).sum()


def m_dreg_looser(model, x, K=1):
    """Computes dreg estimate for log p_\theta(x) for multi-modal vae
    This version is the looser bound---with the average over modalities outside the log
    """
    S = compute_microbatch_split(x, K)
    x_split = zip(*[_x.split(S) for _x in x])
    lw, zss = zip(*[_m_dreg_looser(model, _x, K) for _x in x_split])
    lw = torch.cat(lw, 2)  # concat on batch
    zss = torch.cat(zss, 2)  # concat on batch
    with torch.no_grad():
        grad_wt = (lw - torch.logsumexp(lw, 1, keepdim=True)).exp()
        if zss.requires_grad:
            zss.register_hook(lambda grad: grad_wt.unsqueeze(-1) * grad)
    return (grad_wt * lw).mean(0).sum()


class Test_mmvae_obj:
    @pytest.fixture
    def dataset(self):
        # Create simple small dataset
        data = dict(
            mod1=torch.Tensor([[1.0, 2.0], [4.0, 5.0]]),
            mod2=torch.Tensor([[67.1, 2.3, 3.0], [1.3, 2.0, 3.0]]),
        )
        labels = np.array([0, 1])
        dataset = MultimodalBaseDataset(data, labels)
        return dataset

    @pytest.fixture(params=[True, False])
    def model_config(self, request):
        model_config = MMVAEConfig(
            n_modalities=2,
            latent_dim=5,
            input_dims=dict(mod1=(2,), mod2=(3,)),
            use_likelihood_rescaling=request.param,
            decoder_dist_params=dict(mod1=dict(scale=0.75), mod2=dict(scale=0.75)),
            decoders_dist=dict(mod1="laplace", mod2="laplace"),
            K=5,
        )

        return model_config

    @pytest.fixture
    def model(self, model_config):
        model = MMVAE(model_config)
        return model

    def test_forward(self, model, dataset):
        out = model(dataset, detailed_output=True)

        assert hasattr(out, "qz_xs_detach")

        qz_xs_detach_dict = out.qz_xs_detach
        qz_xs_dict = out.qz_xs

        zss_dict = out.zss
        px_zs_dict = out.recon

        qz_xs_detach = []
        qz_xs = []
        zss = []
        px_zs = []
        keys = list(zss_dict.keys())
        for i, mod in enumerate(keys):
            zss.append(zss_dict[mod])
            qz_xs_detach.append(qz_xs_detach_dict[mod])
            qz_xs.append(qz_xs_dict[mod])
            px_zs.append([])
            for j, r_mod in enumerate(keys):
                recon = px_zs_dict[mod][r_mod]

                px_zs[-1].append(dist.Laplace(recon, 0.75))

        x = [dataset.data[mod] for mod in keys]
        rescale_factors = [model.rescale_factors[mod] for mod in keys]
        test_loss = _m_dreg_looser_test(
            qz_xs_detach, px_zs, zss, model, x, rescale_factors, K=model.K
        )

        assert test_loss == out.loss

        test_iwae = -_m_iwae_test(
            qz_xs, px_zs, zss, model, x, rescale_factors, K=model.K
        )
        model_iwae = model.iwae(qz_xs_dict, zss_dict, out.recon, dataset).loss
        assert test_iwae == model_iwae



import os

import numpy as np
import pytest
import torch
from pythae.models.base import BaseAEConfig
from pythae.models.nn.benchmarks.mnist.convnets import (
    Decoder_Conv_AE_MNIST,
    Encoder_Conv_AE_MNIST,
)
from pythae.models.nn.default_architectures import Encoder_AE_MLP
from torch import nn

from multivae.models.base import BaseMultiVAE, BaseMultiVAEConfig
from multivae.models.nn.default_architectures import (
    Decoder_AE_MLP,
    MultipleHeadJointEncoder,
)


class Test_MultipleHeadJointEncoder:
    @pytest.fixture
    def input_model(self):
        model_config = BaseMultiVAEConfig(n_modalities=2, latent_dim=10)
        config1 = BaseAEConfig(input_dim=(7,), latent_dim=10)
        config2 = BaseAEConfig(input_dim=(3,), latent_dim=10)

        encoders = dict(mod1=Encoder_AE_MLP(config1), mod2=Encoder_AE_MLP(config2))
        data = dict(mod1=torch.ones((3, 7)), mod2=torch.ones((3, 3)))
        return dict(dict_encoders=encoders, args=model_config, data=data)

    def test(self, input_model):
        model = MultipleHeadJointEncoder(**input_model)

        assert type(model.encoders) == nn.ModuleDict
        assert isinstance(model.encoders["mod1"], Encoder_AE_MLP)
        assert model.encoders["mod1"] != input_model["dict_encoders"]["mod1"]
        assert model.latent_dim == input_model["args"].latent_dim

        # forward pass
        outputs = model(input_model["data"])
        assert outputs.embedding.shape == (3, 10)
        assert outputs.log_covariance.shape == (3, 10)



"""In this section, you will find all the models that are currently implemented in `multivae` library
"""

from .auto_model import AutoConfig, AutoModel
from .base import BaseMultiVAE, BaseMultiVAEConfig
from .jmvae import JMVAE, JMVAEConfig
from .jnf import JNF, JNFConfig
from .jnf_dcca import JNFDcca, JNFDccaConfig
from .mmvae import MMVAE, MMVAEConfig
from .mmvaePlus import MMVAEPlus, MMVAEPlusConfig
from .mopoe import MoPoE, MoPoEConfig
from .mvae import MVAE, MVAEConfig
from .mvtcae import MVTCAE, MVTCAEConfig
from .telbo import TELBO, TELBOConfig

__all__ = [
    "BaseMultiVAEConfig",
    "BaseMultiVAE",
    "JMVAEConfig",
    "JMVAE",
    "AutoConfig",
    "AutoModel",
    "JNF",
    "JNFConfig",
    "TELBO",
    "TELBOConfig",
    "MVAE",
    "MVAEConfig",
    "MMVAEConfig",
    "MMVAE",
    "JNFDcca",
    "JNFDccaConfig",
    "MoPoE",
    "MoPoEConfig",
    "MVTCAE",
    "MVTCAEConfig",
    "MMVAEPlusConfig",
    "MMVAEPlus",
]



from typing import Dict, Union

import torch
import torch.distributions as dist
from pythae.models.base.base_utils import ModelOutput
from pythae.models.nn.base_architectures import BaseDecoder, BaseEncoder

from ...data.datasets.base import MultimodalBaseDataset
from ..joint_models import BaseJointModel
from .telbo_config import TELBOConfig


class TELBO(BaseJointModel):

    """
    The Triple ELBO VAE model.

    Args:

        model_config (TELBOConfig): An instance of TELBOConfig in which any model's parameters is
            made available.

        encoders (Dict[str, ~pythae.models.nn.base_architectures.BaseEncoder]): A dictionary containing
            the modalities names and the encoders for each modality. Each encoder is an instance of
            Pythae's BaseEncoder. Default: None.

        decoder (Dict[str, ~pythae.models.nn.base_architectures.BaseDecoder]): A dictionary containing
            the modalities names and the decoders for each modality. Each decoder is an instance of
            Pythae's BaseDecoder.

        joint_encoder (~pythae.models.nn.base_architectures.BaseEncoder) : An instance of BaseEncoder that takes all the modalities as an input.
            If none is provided, one is created from the unimodal encoders. Default : None.


    """

    def __init__(
        self,
        model_config: TELBOConfig,
        encoders: Dict[str, BaseEncoder] = None,
        decoders: Dict[str, BaseDecoder] = None,
        joint_encoder: Union[BaseEncoder, None] = None,
        **kwargs,
    ):
        super().__init__(model_config, encoders, decoders, joint_encoder, **kwargs)

        self.model_name = "TELBO"
        self.warmup = model_config.warmup
        self.reset_optimizer_epochs = [self.warmup]

        if model_config.lambda_factors is None:
            self.lambda_factors = self.rescale_factors
        else:
            self.lambda_factors = model_config.lambda_factors
        if model_config.gamma_factors is None:
            self.gamma_factors = self.rescale_factors
        else:
            self.gamma_factors = model_config.gamma_factors

    def _set_torch_no_grad_on_joint_vae(self):
        """Function used to freeze the parameters of the joint
        encoder and decoders after the warmup."""
        self.joint_encoder.requires_grad_(False)
        self.decoders.requires_grad_(False)

    def forward(self, inputs: MultimodalBaseDataset, **kwargs):
        epoch = kwargs.pop("epoch", 1)

        # First compute the joint ELBO
        joint_output = self.joint_encoder(inputs.data)
        mu, log_var = joint_output.embedding, joint_output.log_covariance

        sigma = torch.exp(0.5 * log_var)
        qz_xy = dist.Normal(mu, sigma)
        z_joint = qz_xy.rsample()

        recon_loss = 0

        # Decode in each modality
        len_batch = 0
        for mod in self.decoders:
            x_mod = inputs.data[mod]
            len_batch = len(x_mod)
            recon_mod = self.decoders[mod](z_joint).reconstruction
            recon_loss += (
                -self.recon_log_probs[mod](recon_mod, x_mod) * self.lambda_factors[mod]
            ).sum()

        # Compute the KLD to the prior
        KLD = -0.5 * torch.sum(1 + log_var - mu.pow(2) - log_var.exp())

        if epoch <= self.warmup:
            return ModelOutput(
                recon_loss=recon_loss / len_batch,
                KLD=KLD / len_batch,
                loss=(recon_loss + KLD) / len_batch,
                metrics=dict(kld_joint=KLD, recon_joint=recon_loss / len_batch),
            )

        else:
            self._set_torch_no_grad_on_joint_vae()
            mod_elbos = {}
            loss = 0

            for mod in self.encoders:
                mod_output = self.encoders[mod](inputs.data[mod])
                mod_mu, mod_log_var = mod_output.embedding, mod_output.log_covariance

                mod_sigma = torch.exp(0.5 * mod_log_var)
                qz_x0 = dist.Normal(mod_mu, mod_sigma)
                mod_z = qz_x0.rsample()

                mod_recon = self.decoders[mod](mod_z).reconstruction
                mod_recon_loss = (
                    -self.recon_log_probs[mod](mod_recon, inputs.data[mod])
                    * self.gamma_factors[mod]
                )

                mod_kld = -0.5 * torch.sum(
                    1 + log_var - mod_mu.pow(2) - mod_log_var.exp()
                )
                mod_elbos[mod] = mod_recon_loss.sum() + mod_kld
                loss += mod_recon_loss.sum() + mod_kld

            return ModelOutput(loss=loss / len_batch, metrics=mod_elbos)

    def encode(
        self,
        inputs: MultimodalBaseDataset,
        cond_mod: Union[list, str] = "all",
        N: int = 1,
        **kwargs,
    ) -> ModelOutput:
        self.eval()

        if type(cond_mod) == list and len(cond_mod) == 1:
            cond_mod = cond_mod[0]

        if cond_mod == "all" or (
            type(cond_mod) == list and len(cond_mod) == self.n_modalities
        ):
            output = self.joint_encoder(inputs.data)
            sample_shape = [] if N == 1 else [N]
            z = dist.Normal(
                output.embedding, torch.exp(0.5 * output.log_covariance)
            ).rsample(sample_shape)
            if N > 1 and kwargs.pop("flatten", False):
                N, l, d = z.shape
                z = z.reshape(l * N, d)
            return ModelOutput(z=z, one_latent_space=True)

        if type(cond_mod) == list and len(cond_mod) != 1:
            raise AttributeError(
                "Conditioning on a subset containing more than one modality "
                "is not yet implemented."
            )

        if cond_mod in self.modalities_name:
            output = self.encoders[cond_mod](inputs.data[cond_mod])
            sample_shape = [] if N == 1 else [N]

            z = dist.Normal(
                output.embedding, torch.exp(0.5 * output.log_covariance)
            ).rsample(sample_shape)

            if N > 1 and kwargs.pop("flatten", False):
                z = z.reshape(-1, self.latent_dim)

            return ModelOutput(z=z, one_latent_space=True)
        else:
            raise AttributeError(
                f"Modality of name {cond_mod} not handled. The"
                f" modalities that can be encoded are {list(self.encoders.keys())}"
            )



from typing import Union

import numpy as np
import torch
import torch.distributions as dist
from pythae.models.base.base_utils import ModelOutput

from multivae.data.datasets.base import IncompleteDataset, MultimodalBaseDataset

from ..base import BaseMultiVAE
from .mvtcae_config import MVTCAEConfig


class MVTCAE(BaseMultiVAE):

    """

    Implementation for 'Multi-View Representation Learning via Total Correlation Objective'.
    Hwang et al, 2021.

    This code is heavily based on the official implementation that can be found here :
    https://github.com/gr8joo/MVTCAE.


    """

    def __init__(
        self, model_config: MVTCAEConfig, encoders: dict = None, decoders: dict = None
    ):
        super().__init__(model_config, encoders, decoders)

        self.alpha = model_config.alpha
        self.beta = model_config.beta
        self.model_name = "MVTCAE"

    def reparameterize(self, mu, logvar):
        std = logvar.mul(0.5).exp_()
        z = dist.Normal(mu, std).rsample()
        return z

    def forward(self, inputs: MultimodalBaseDataset, **kwargs) -> ModelOutput:
        # Compute latents parameters for all subsets
        latents = self.inference(inputs)
        results = dict()

        # Sample from the joint posterior
        joint_mu, joint_logvar = latents["joint"][0], latents["joint"][1]
        shared_embeddings = self.reparameterize(joint_mu, joint_logvar)
        ndata = len(shared_embeddings)
        joint_kld = -0.5 * torch.sum(
            1 - joint_logvar.exp() - joint_mu.pow(2) + joint_logvar
        )
        assert not torch.isnan(joint_kld)
        results["joint_divergence"] = joint_kld

        # Compute the reconstruction losses for each modality
        loss_rec = 0
        for m_key in self.encoders.keys():
            # reconstruct this modality from the shared embeddings representation
            recon = self.decoders[m_key](shared_embeddings).reconstruction
            m_rec = (
                -self.recon_log_probs[m_key](recon, inputs.data[m_key])
                * self.rescale_factors[m_key]
            )
            m_rec = m_rec.reshape(m_rec.size(0), -1).sum(-1)

            # Keep only the available samples
            if hasattr(inputs, "masks"):
                m_rec = inputs.masks[m_key].float() * m_rec

            results[m_key] = m_rec.sum()
            loss_rec += m_rec.sum()
        assert not torch.isnan(loss_rec)

        latent_modalities = latents["modalities"]
        kld_losses = 0.0
        for m_key in latent_modalities.keys():
            o = latent_modalities[m_key]
            mu, logvar = o.embedding, o.log_covariance
            results["kld_" + m_key] = -0.5 * (
                1
                - joint_logvar.exp() / logvar.exp()
                - (joint_mu - mu).pow(2) / logvar.exp()
                + joint_logvar
                - logvar
            ).reshape(mu.size(0), -1).sum(-1)

            # Keep only the available samples
            if hasattr(inputs, "masks"):
                results["kld_" + m_key][(1 - inputs.masks[m_key].int()).bool()] = 0

            results["kld_" + m_key] = results["kld_" + m_key].sum()

            kld_losses += results["kld_" + m_key].sum()
        assert not torch.isnan(kld_losses)

        rec_weight = (self.n_modalities - self.alpha) / self.n_modalities
        cvib_weight = self.alpha / self.n_modalities  # 1/6
        vib_weight = 1 - self.alpha  # 0.1

        kld_weighted = cvib_weight * kld_losses + vib_weight * joint_kld
        total_loss = rec_weight * loss_rec + self.beta * kld_weighted

        return ModelOutput(loss=total_loss / ndata, metrics=results)

    def modality_encode(
        self, inputs: Union[MultimodalBaseDataset, IncompleteDataset], **kwargs
    ):
        """Computes for each modality, the parameters mu and logvar of the
        unimodal posterior.

        Args:
            inputs (MultimodalBaseDataset): The data to encode.

        Returns:
            dict: Containing for each modality the encoder output.
        """
        encoders_outputs = dict()
        for m, m_key in enumerate(inputs.data.keys()):
            input_modality = inputs.data[m_key]
            output = self.encoders[m_key](input_modality)
            # For unavailable samples, set the log-variance to infty so that they don't contribute to the
            # product of experts
            if hasattr(inputs, "masks"):
                output.log_covariance[
                    (1 - inputs.masks[m_key].int()).bool()
                ] = torch.inf
            encoders_outputs[m_key] = output

        return encoders_outputs

    def poe(self, mu, logvar, eps=1e-8):
        var = torch.exp(logvar) + eps
        # precision of i-th Gaussian expert at point x
        T = 1.0 / var
        pd_mu = torch.sum(mu * T, dim=0) / torch.sum(T, dim=0)
        pd_var = 1.0 / torch.sum(T, dim=0)
        pd_logvar = torch.log(pd_var)
        return pd_mu, pd_logvar

    def ivw_fusion(self, mus: torch.Tensor, logvars: torch.Tensor, weights=None):
        mu_poe, logvar_poe = self.poe(mus, logvars)
        return [mu_poe, logvar_poe]

    def _filter_inputs_with_masks(
        self, inputs: IncompleteDataset, subset: Union[list, tuple]
    ):
        """
        Returns a filtered dataset containing only the samples that are available
        in all the modalities contained in subset.
        The dataset that is returned only contains the modalities in subset.
        """

        filter = torch.tensor(
            True,
        ).to(inputs.masks[subset[0]].device)
        for mod in subset:
            filter = torch.logical_and(filter, inputs.masks[mod])

        filtered_inputs = MultimodalBaseDataset(
            data={k: inputs.data[k][filter] for k in subset},
        )
        return filtered_inputs, filter

    def inference(self, inputs: MultimodalBaseDataset, **kwargs):
        """
        This function takes all the modalities contained in inputs
        and compute the product of experts of the modalities encoders.

        Args:
            inputs (MultimodalBaseDataset): The data.

        Returns:
            dict : Contains the modalities' encoders parameters and the poe parameters.
        """

        latents = dict()
        enc_mods = self.modality_encode(inputs)
        latents["modalities"] = enc_mods

        device = inputs.data[list(inputs.data.keys())[0]].device
        mus = torch.Tensor().to(device)
        logvars = torch.Tensor().to(device)
        mods = list(inputs.data.keys())

        for m, mod in enumerate(mods):
            mus_mod = enc_mods[mod].embedding
            log_vars_mod = enc_mods[mod].log_covariance

            mus = torch.cat((mus, mus_mod.unsqueeze(0)), dim=0)

            logvars = torch.cat((logvars, log_vars_mod.unsqueeze(0)), dim=0)

        # Case with only one sample : adapt the shape
        if len(mus.shape) == 2:
            mus = mus.unsqueeze(1)
            logvars = logvars.unsqueeze(1)

        joint_mu, joint_logvar = self.ivw_fusion(mus, logvars)

        latents["joint"] = [joint_mu, joint_logvar]
        return latents

    def encode(
        self,
        inputs: MultimodalBaseDataset,
        cond_mod: Union[list, str] = "all",
        N: int = 1,
        **kwargs,
    ) -> ModelOutput:
        cond_mod = super().encode(inputs, cond_mod, N, **kwargs).cond_mod

        # Only keep the relevant modalities for prediction
        cond_inputs = MultimodalBaseDataset(
            data={k: inputs.data[k] for k in cond_mod},
        )

        latents_subsets = self.inference(cond_inputs)
        mu, log_var = latents_subsets["joint"]
        sample_shape = [N] if N > 1 else []

        return_mean = kwargs.pop("return_mean", False)
        if return_mean:
            z = torch.stack([mu] * N) if N > 1 else mu
        else:
            z = dist.Normal(mu, torch.exp(0.5 * log_var)).rsample(sample_shape)
        flatten = kwargs.pop("flatten", False)
        if flatten:
            z = z.reshape(-1, self.latent_dim)

        return ModelOutput(z=z, one_latent_space=True)

    def compute_joint_nll(
        self,
        inputs: Union[MultimodalBaseDataset, IncompleteDataset],
        K: int = 1000,
        batch_size_K: int = 100,
    ):
        self.eval()

        # Compute the parameters of the joint posterior
        mu, log_var = self.inference(inputs)["joint"]

        sigma = torch.exp(0.5 * log_var)
        qz_xy = dist.Normal(mu, sigma)
        # And sample from the posterior
        z_joint = qz_xy.rsample([K])  # shape K x n_data x latent_dim
        z_joint = z_joint.permute(1, 0, 2)
        n_data, _, latent_dim = z_joint.shape

        # Then iter on each datapoint to compute the iwae estimate of ln(p(x))
        ll = 0
        for i in range(n_data):
            start_idx = 0
            stop_idx = min(start_idx + batch_size_K, K)
            lnpxs = []
            while start_idx < stop_idx:
                latents = z_joint[i][start_idx:stop_idx]

                # Compute p(x_m|z) for z in latents and for each modality m
                lpx_zs = 0  # ln(p(x,y|z))
                for mod in inputs.data:
                    decoder = self.decoders[mod]
                    recon = decoder(latents)[
                        "reconstruction"
                    ]  # (batch_size_K, nb_channels, w, h)
                    x_m = inputs.data[mod][i]  # (nb_channels, w, h)

                    lpx_zs += (
                        self.recon_log_probs[mod](
                            recon, torch.stack([x_m] * len(recon))
                        )
                        .reshape(recon.size(0), -1)
                        .sum(-1)
                    )

                # Compute ln(p(z))
                prior = dist.Normal(0, 1)
                lpz = prior.log_prob(latents).sum(dim=-1)

                # Compute posteriors -ln(q(z|x,y))
                qz_xy = dist.Normal(mu[i], sigma[i])
                lqz_xy = qz_xy.log_prob(latents).sum(dim=-1)

                ln_px = torch.logsumexp(lpx_zs + lpz - lqz_xy, dim=0)
                lnpxs.append(ln_px)

                # next batch
                start_idx += batch_size_K
                stop_idx = min(stop_idx + batch_size_K, K)

            ll += torch.logsumexp(torch.Tensor(lnpxs), dim=0) - np.log(K)

        return -ll



import os

import numpy as np
import pytest
import torch

from multivae.data.datasets import MultimodalBaseDataset
from multivae.data.datasets.utils import ResampleDataset, adapt_shape


@pytest.fixture
def dummy_data():
    return dict(
        mod1=torch.randn(6, 1, 28, 28),
        mod2=torch.randn(6, 3, 28),
        mod3=torch.randn(6, 3, 28, 28),
        mod4=torch.randn(6, 2, 28, 28),
        mod5=torch.randn(6, 20, 28, 28),
        mod6=torch.randn(6, 28),
    )


class TestUtils:
    def test_adapt_shape(self, dummy_data):
        out, _ = adapt_shape(dummy_data)
        assert out.keys() == out.keys()
        assert all([out[k].shape[1] == 3 for k in dummy_data.keys()])

        new_dummy_data = dict(wrong_mod=torch.randn(10, 2, 2, 4, 2))
        with pytest.raises(AttributeError):
            _ = adapt_shape(new_dummy_data)

    def test_resample_dataset(self, dummy_data):
        dataset = ResampleDataset(
            MultimodalBaseDataset(dummy_data, labels=torch.ones(6)),
            transform=lambda x: x,
        )
        assert all(
            [
                dataset[torch.tensor(0)].data[m].shape == dummy_data[m].shape[1:]
                for m in dummy_data.keys()
            ]
        )



from pythae.models.base.base_utils import ModelOutput

from multivae.data import MultimodalBaseDataset

from ..base.evaluator_class import Evaluator
from .likelihoods_config import LikelihoodsEvaluatorConfig

try:
    from tqdm import tqdm
except:
    tqdm = lambda x: x

from multivae.data.utils import set_inputs_to_device


class LikelihoodsEvaluator(Evaluator):
    """
    Class for computing likelihood metrics.

    Args:

        model (BaseMultiVAE) : The model to evaluate.
        test_dataset (MultimodalBaseDataset) : The dataset to use for computing the metrics.
        output (str) : The folder path to save metrics. The metrics will be saved in a metrics.txt
            file.
        eval_config (EvaluatorConfig) : The configuration class to specify parameters for the
            evaluation.
    """

    def __init__(
        self, model, test_dataset, output=None, eval_config=LikelihoodsEvaluatorConfig()
    ) -> None:
        super().__init__(model, test_dataset, output, eval_config)
        self.num_samples = eval_config.num_samples
        self.batch_size_k = eval_config.batch_size_k
        self.unified = eval_config.unified_implementation

    def eval(self):
        self.joint_nll()
        self.log_to_wandb()
        return ModelOutput(**self.metrics)

    def joint_nll(self):
        ll = 0
        for batch in tqdm(self.test_loader):
            batch = set_inputs_to_device(batch, self.device)
            if self.unified or (not hasattr(self.model, "compute_joint_nll_paper")):
                ll += self.model.compute_joint_nll(
                    batch, self.num_samples, self.batch_size_k
                )
            else:
                self.logger.info("Using the paper version of the joint nll.")
                ll += self.model.compute_joint_nll_paper(
                    batch, self.num_samples, self.batch_size_k
                )

        joint_nll = ll / len(self.test_loader.dataset)
        self.logger.info(f"Joint likelihood : {str(joint_nll)}")
        self.metrics["joint_likelihood"] = joint_nll
        return joint_nll

    def joint_nll_from_subset(self, subset):
        if hasattr(self.model, "compute_joint_nll_from_subset_encoding"):
            ll = 0
            nb_batch = 0
            for batch in self.test_loader:
                batch = set_inputs_to_device(batch, self.device)
                ll += self.model.compute_joint_nll_from_subset_encoding(
                    subset, batch, self.num_samples, self.batch_size_k
                )
                nb_batch += 1

            joint_nll = ll / self.n_data
            self.logger.info(
                f"Joint likelihood from subset {subset} : {str(joint_nll)}"
            )
            self.metrics[f"Joint likelihood from subset {subset}"] = joint_nll
            return joint_nll
        else:
            return None



from typing import Literal

from pydantic.dataclasses import dataclass

from ..base.evaluator_config import EvaluatorConfig


@dataclass
class ClusteringConfig(EvaluatorConfig):

    """

    Config class for the clustering module.

    Args:
        batch_size (int) : The batch size to use in the evaluation. Default to 512
        wandb_path (str) : The user can provide the path of the wandb run with a
            format 'entity/projet_name/run_id' where the metrics should be logged.
            For an existing run (the training run), the info can be found in the training dir
            (in wandb_info.json)
            at the end of training (if wandb was used) or on the hugging_face webpage of the run.
            Otherwise the user can create a new wandb run and get the path with :

               .. code-block:: python

                    >>> import wandb
                    >>> run = wandb.init(entity = your_entity, project=your_project)
                    >>> wandb_path = run.path

            If None are provided, the metrics are not logged on wandb.
            Default to None.
        clustering_method (Literal['kmeans']) :  The method to use to cluster. Default to 'kmeans'
        n_clusters (int) :the number of clusters. Default to 10.
        number_of_runs (int) : When computing accuracies, how many runs of clustering to perform to
        to compute the average accuracies. Default to 20.
        num-samples_for_fit (int) : Number of training samples to use to fit the clusters. If None,
            uses all the samples. Default to None.
        use_mean (bool) : Whether to use a sample or the mean of the encoding distribution as the
            representative embedding. Default to True.

    """

    clustering_method: Literal["kmeans"] = "kmeans"
    n_clusters: int = 10
    number_of_runs: int = 20
    num_samples_for_fit: int = None
    use_mean: bool = True



from typing import Literal

from torchvision import transforms
from torchvision.datasets import CelebA

from .base import DatasetOutput, MultimodalBaseDataset


class CelebAttr(MultimodalBaseDataset):
    
    def __init__(self, root : str, split :str, transform = None, target_transform = None,attributes : Literal['18','40']='18',download = False): 
        
        self.root = root
        
        if transform is None:
            transform = transforms.Compose([transforms.Resize(64),
                                          transforms.CenterCrop(64),
                                          transforms.ToTensor()])
        self.transform = transform
        
        self.torchvision_dataset = CelebA(root=root,
                                          split=split,
                                          target_type='attr',
                                          transform = transform,
                                          download=download
                                          )
        
        if attributes == '18' :
            self.attributes_to_keep = [4, 5, 8, 9, 11, 12, 15, 17, 18, 20, 21, 22, 26, 28, 31, 32, 33, 35]
        else:
            self.attributes_to_keep = range(40)
            
        self.attr_to_idx  = {'Sideburns': 30, 'Black_Hair': 8, 'Wavy_Hair': 33, 'Young': 39, 'Heavy_Makeup': 18, 
                    'Blond_Hair': 9, 'Attractive': 2, '5_o_Clock_Shadow': 0, 'Wearing_Necktie': 38, 
                    'Blurry': 10, 'Double_Chin': 14, 'Brown_Hair': 11, 'Mouth_Slightly_Open': 21, 
                    'Goatee': 16, 'Bald': 4, 'Pointy_Nose': 27, 'Gray_Hair': 17, 'Pale_Skin': 26, 
                    'Arched_Eyebrows': 1, 'Wearing_Hat': 35, 'Receding_Hairline': 28, 'Straight_Hair': 32, 
                    'Big_Nose': 7, 'Rosy_Cheeks': 29, 'Oval_Face': 25, 'Bangs': 5, 'Male': 20, 'Mustache': 22, 
                    'High_Cheekbones': 19, 'No_Beard': 24, 'Eyeglasses': 15, 'Bags_Under_Eyes': 3, 
                    'Wearing_Necklace': 37, 'Wearing_Lipstick': 36, 'Big_Lips': 6, 'Narrow_Eyes': 23, 
                    'Chubby': 13, 'Smiling': 31, 'Bushy_Eyebrows': 12, 'Wearing_Earrings': 34}
        
        self.idx_to_attr = {v:k for k, v in self.attr_to_idx.items()}

    def __getitem__(self, index):
        
        img,target = self.torchvision_dataset[index]
        
        return DatasetOutput(
            data = dict(image = img,
                        attributes = target[self.attributes_to_keep]),
            labels = target[self.attributes_to_keep]
            
        )
        
    def __len__(self):
        return self.torchvision_dataset.__len__()
        



from pydantic.dataclasses import dataclass

from ..base import BaseSamplerConfig


@dataclass
class GaussianMixtureSamplerConfig(BaseSamplerConfig):
    """Gaussian mixture sampler config class.

    Parameters:

        n_components (int): The number of Gaussians in the mixture. Default to 10
    """

    n_components: int = 10



from itertools import combinations
from typing import Union

import numpy as np
import torch
import torch.distributions as dist
from numpy.random import choice
from pythae.models.base.base_utils import ModelOutput

from multivae.data.datasets.base import IncompleteDataset, MultimodalBaseDataset

from ..base import BaseMultiVAE
from .mvae_config import MVAEConfig


class MVAE(BaseMultiVAE):
    """
    The Multi-modal VAE model.

    Args:
        model_config (MVAEConfig): An instance of MVAEConfig in which any model's
            parameters is made available.

        encoders (Dict[str, ~pythae.models.nn.base_architectures.BaseEncoder]): A dictionary containing
            the modalities names and the encoders for each modality. Each encoder is an instance of
            Pythae's BaseEncoder. Default: None.

        decoder (Dict[str, ~pythae.models.nn.base_architectures.BaseDecoder]): A dictionary containing
            the modalities names and the decoders for each modality. Each decoder is an instance of
            Pythae's BaseDecoder.
    """

    def __init__(
        self, model_config: MVAEConfig, encoders: dict = None, decoders: dict = None
    ):
        super().__init__(model_config, encoders, decoders)

        self.subsampling = model_config.use_subsampling
        self.k = model_config.k
        if self.n_modalities <= 2:
            self.k = 0
        self.set_subsets()
        self.warmup = model_config.warmup
        self.beta = model_config.beta
        self.model_name = "MVAE"

    def set_subsets(self):
        self.subsets = []
        for i in range(2, self.n_modalities):
            self.subsets += combinations(list(self.encoders.keys()), r=i)

    def poe(self, mus_list, log_vars_list):
        mus = mus_list.copy()
        log_vars = log_vars_list.copy()

        # Add the prior to the product of experts
        mus.append(torch.zeros_like(mus[0]))
        log_vars.append(torch.zeros_like(log_vars[0]))

        # Compute the joint posterior
        lnT = torch.stack([-l for l in log_vars])  # Compute the inverse of variances
        lnV = -torch.logsumexp(lnT, dim=0)  # variances of the product of expert
        mus = torch.stack(mus)
        joint_mu = (torch.exp(lnT) * mus).sum(dim=0) * torch.exp(lnV)

        return joint_mu, lnV

    # def poe(self, mus_list, logvar_list, eps=1e-8):

    # ORIGINAL VERSION BUT LESS STABLE

    #     mus = mus_list.copy()
    #     log_vars = logvar_list.copy()

    #     # Add the prior to the product of experts
    #     mus.append(torch.zeros_like(mus[0]))
    #     log_vars.append(torch.zeros_like(log_vars[0]))

    #     mus = torch.stack(mus)
    #     logvars = torch.stack(log_vars)
    #     var       = torch.exp(logvars) + eps
    #     # precision of i-th Gaussian expert at point x
    #     T         = 1. / (var + eps)
    #     pd_mu     = torch.sum(mus * T, dim=0) / torch.sum(T, dim=0)
    #     pd_var    = 1. / torch.sum(T, dim=0)
    #     pd_logvar = torch.log(pd_var + eps)
    #     return pd_mu, pd_logvar

    def compute_mu_log_var_subset(self, inputs: MultimodalBaseDataset, subset: list):
        """Computes the parameters of the posterior when conditioning on
        the modalities contained in subset."""
        mus_sub = []
        log_vars_sub = []
        for mod in self.encoders:
            if mod in subset:
                output_mod = self.encoders[mod](inputs.data[mod])
                mu_mod, log_var_mod = output_mod.embedding, output_mod.log_covariance
                if hasattr(inputs, "masks"):
                    log_var_mod[
                        (1 - inputs.masks[mod].int()).bool().flatten()
                    ] = torch.inf

                mus_sub.append(mu_mod)
                log_vars_sub.append(log_var_mod)
        sub_mu, sub_logvar = self.poe(mus_sub, log_vars_sub)
        return sub_mu, sub_logvar

    def _compute_elbo_subset(
        self, inputs: MultimodalBaseDataset, subset: list, beta: float
    ):
        sub_mu, sub_logvar = self.compute_mu_log_var_subset(inputs, subset)
        sub_std = torch.exp(0.5 * sub_logvar)
        z = dist.Normal(sub_mu, sub_std).rsample()
        elbo_sub = 0
        for mod in self.decoders:
            if mod in subset:
                recon = self.decoders[mod](z).reconstruction
                recon_mod = (
                    -(
                        self.recon_log_probs[mod](recon, inputs.data[mod])
                        * self.rescale_factors[mod]
                    )
                    .reshape(recon.size(0), -1)
                    .sum(-1)
                )

                if hasattr(inputs, "masks"):
                    recon_mod = inputs.masks[mod].float() * recon_mod
                elbo_sub += recon_mod.sum()

        recon = elbo_sub
        KLD = -0.5 * torch.sum(1 + sub_logvar - sub_mu.pow(2) - sub_logvar.exp())
        elbo_sub += KLD * beta

        return elbo_sub / len(sub_mu), KLD / len(sub_mu), recon / len(sub_mu)

    def _filter_inputs_with_masks(
        self, inputs: IncompleteDataset, subset: Union[list, tuple]
    ):
        """Returns a filtered dataset containing only the samples that are available
        in at least one of the modalities contained in subset."""

        filter = torch.tensor(
            False,
        ).to(inputs.masks[subset[0]].device)
        for mod in subset:
            filter = torch.logical_or(filter, inputs.masks[mod])

        filtered_inputs = {}
        filtered_masks = {}

        for mod in subset:
            filtered_inputs[mod] = inputs.data[mod][filter]
            filtered_masks[mod] = inputs.masks[mod][filter]

        filtered_dataset = IncompleteDataset(data=filtered_inputs, masks=filtered_masks)
        return filtered_dataset, filter

    def forward(
        self, inputs: Union[MultimodalBaseDataset, IncompleteDataset], **kwargs
    ):
        """The main function of the model that computes the loss and some monitoring metrics.
        One of the advantages of MVAE is that we can train with incomplete data.

        Args:
            inputs (MultimodalBaseDataset): The data. It can be an instance of IncompleteDataset
                which contains a field masks for weakly supervised learning.
                masks is a dictionary indicating which datasamples are missing in each of the modalities.
                For each modality, a boolean tensor indicates which samples are available. (The non
                available samples are assumed to be replaced with zero values in the multimodal dataset entry.)
        """

        epoch = kwargs.pop("epoch", 1)
        batch_ratio = kwargs.pop("batch_ratio", 0)
        if epoch >= self.warmup:
            beta = 1 * self.beta
        else:
            beta = (epoch - 1 + batch_ratio) / self.warmup * self.beta
        total_loss = 0
        metrics = {}
        # Collect all the subsets
        subsets = []
        # Add the joint subset
        subsets.append([m for m in self.encoders])
        if self.subsampling:
            # Add the unimodal subsets
            subsets.extend([[m] for m in self.encoders])
            # Add random subsets
            if self.k > 0:
                random_idx = choice(
                    np.arange(len(self.subsets)), size=self.k, replace=False
                )
                for id in random_idx:
                    subsets.append(self.subsets[id])

        for s in subsets:
            if hasattr(inputs, "masks"):
                filtered_inputs, filter = self._filter_inputs_with_masks(inputs, s)
                not_all_samples_missing = torch.any(filter)
            else:
                filtered_inputs = inputs
                not_all_samples_missing = True

            if not_all_samples_missing:
                subset_elbo, subset_kld, subset_recon = self._compute_elbo_subset(
                    filtered_inputs, s, beta
                )
            else:
                subset_elbo = 0
            total_loss += subset_elbo
            metrics["_".join(sorted(s))] = subset_elbo
            metrics["beta"] = beta
            metrics["kld" + "_".join(sorted(s))] = subset_kld
            metrics["recon" + "_".join(sorted(s))] = subset_recon

        return ModelOutput(loss=total_loss, metrics=metrics)

    def encode(
        self,
        inputs: Union[MultimodalBaseDataset, IncompleteDataset],
        cond_mod: Union[list, str] = "all",
        N: int = 1,
        **kwargs,
    ):
        cond_mod = super().encode(inputs, cond_mod, N, **kwargs).cond_mod

        sub_mu, sub_logvar = self.compute_mu_log_var_subset(inputs, cond_mod)
        sub_std = torch.exp(0.5 * sub_logvar)
        sample_shape = [N] if N > 1 else []

        return_mean = kwargs.pop("return_mean", False)

        if return_mean:
            z = torch.stack([sub_mu] * N) if N > 1 else sub_mu
        else:
            z = dist.Normal(sub_mu, sub_std).rsample(sample_shape)
        flatten = kwargs.pop("flatten", False)
        if flatten:
            z = z.reshape(-1, self.latent_dim)

        return ModelOutput(z=z, one_latent_space=True)

    def compute_joint_nll(
        self,
        inputs: Union[MultimodalBaseDataset, IncompleteDataset],
        K: int = 1000,
        batch_size_K: int = 100,
    ):
        """Computes the joint_negative_nll for a batch of inputs."""

        # iter on each datapoint to compute the iwae estimate of ln(p(x))
        ll = 0
        n_data = len(inputs.data[list(inputs.data.keys())[0]])
        for i in range(n_data):
            start_idx = 0
            stop_idx = min(start_idx + batch_size_K, K)
            lnpxs = []

            # Compute the parameters of the joint posterior
            mu, log_var = self.compute_mu_log_var_subset(
                MultimodalBaseDataset(
                    data={k: inputs.data[k][i].unsqueeze(0) for k in inputs.data}
                ),
                list(self.encoders.keys()),
            )

            sigma = torch.exp(0.5 * log_var)
            qz_xy = dist.Normal(mu, sigma)

            # And sample from the posterior
            z_joint = qz_xy.rsample([K]).squeeze()  # shape K x latent_dim

            while start_idx < stop_idx:
                latents = z_joint[start_idx:stop_idx]

                # Compute p(x_m|z) for z in latents and for each modality m
                lpx_zs = 0  # ln(p(x,y|z))
                for mod in inputs.data:
                    decoder = self.decoders[mod]
                    recon = decoder(latents)[
                        "reconstruction"
                    ]  # (batch_size_K, nb_channels, w, h)
                    x_m = inputs.data[mod][i]  # (nb_channels, w, h)

                    lpx_zs += (
                        self.recon_log_probs[mod](
                            recon, torch.stack([x_m] * len(recon))
                        )
                        .reshape(recon.size(0), -1)
                        .sum(-1)
                    )

                # Compute ln(p(z))
                prior = dist.Normal(0, 1)
                lpz = prior.log_prob(latents).sum(dim=-1)

                # Compute posteriors -ln(q(z|x,y))
                qz_xy = dist.Normal(mu.squeeze(), sigma.squeeze())
                lqz_xy = qz_xy.log_prob(latents).sum(dim=-1)

                ln_px = torch.logsumexp(lpx_zs + lpz - lqz_xy, dim=0)
                lnpxs.append(ln_px)

                # next batch
                start_idx += batch_size_K
                stop_idx = min(stop_idx + batch_size_K, K)

            ll += torch.logsumexp(torch.Tensor(lnpxs), dim=0) - np.log(K)

        return -ll



"""
Implementation of the Variational Mixture-of-Experts Autoencoder model from the paper "Variational Mixture-of-Experts Autoencoders for
Multi-Modal Deep Generative Models"
(https://arxiv.org/abs/1911.03393)

"""

from .mmvae_config import MMVAEConfig
from .mmvae_model import MMVAE

__all__ = ["MMVAE", "MMVAEConfig"]



"""

Sampler fitting an Inverse Autoregressive Flow in the multimodal variational autoencoder's latent space.
This class is simply adapted from Pythae's iaf_sampler for the multimodal case. 


"""

from .iaf_sampler import IAFSampler
from .iaf_sampler_config import IAFSamplerConfig

__all__ = ["IAFSampler", "IAFSamplerConfig"]



from math import ceil, floor

import torch
import torch.nn.functional as F
from torch.utils.data import Dataset
from torchvision.transforms import ToTensor
from torchvision.utils import save_image


class ResampleDataset(Dataset):
    """
    Dataset which resamples a given dataset. From torchnet's ResampleDataset.
    https://tnt.readthedocs.io/en/latest/_modules/torchnet/dataset/resampledataset.html


    Args:
        dataset (Dataset): Dataset to be resampled.
        sampler (function, optional): Function used for sampling. `idx`th
            sample is returned by `dataset[sampler(dataset, idx)]`. By default
            `sampler(dataset, idx)` is the identity, simply returning `idx`.
            `sampler(dataset, idx)` must return an index in the range
            acceptable for the underlying `dataset`.
        size (int, optional): Desired size of the dataset after resampling. By
            default, the new dataset will have the same size as the underlying
            one.

    """

    def __init__(self, dataset, sampler=lambda ds, idx: idx, size=None, transform=None):
        super(ResampleDataset, self).__init__()
        self.dataset = dataset
        self.sampler = sampler
        self.size = size
        self.transform = transform

    def __len__(self):
        return (self.size and self.size > 0) and self.size or len(self.dataset)

    def __getitem__(self, idx):
        idx = self.sampler(self.dataset, idx)

        if torch.min(idx) < 0 or torch.max(idx) >= len(self.dataset):
            raise IndexError("out of range")

        if self.transform is not None:
            return self.transform(self.dataset[idx])
        else:
            return self.dataset[idx]


def adapt_shape(data):
    """
    Adapts the shape of the data for visualization. The output dictionary contains the same data
    but resized to be of the shape (n_data, ch=3, h, w) with h and w being the largest height and
    width accross data modalities.
    """
    # First, add dimensions if some are missing and adjust the number of channels
    for m in data:
        if len(data[m].shape) == 1:  # (n_data,)
            data[m] = data[m].unsqueeze(1)
        if len(data[m].shape) == 2:  # (n_data, n)
            data[m] = data[m].unsqueeze(1)
        if len(data[m].shape) == 3:  # (n-data, n, m)
            data[m] = data[m].unsqueeze(1)
        if len(data[m].shape) == 4:
            if data[m].shape[1] == 1:
                # Add channels to have 3 channels
                data[m] = torch.cat([data[m] for _ in range(3)], dim=1)
            elif data[m].shape[1] == 2:
                n, ch, h, w = data[m].shape
                data[m] = torch.cat([data[m], torch.zeros(n, 1, h, w)], dim=1)
            else:
                data[m] = data[m][:, :3, :, :]
        else:
            raise AttributeError("Can't visualize data with more than 3 dimensions")

    h = max([data[m].shape[2] for m in data])
    w = max([data[m].shape[3] for m in data])
    for m in data:
        hm, wm = data[m].shape[2:]
        data[m] = F.pad(
            data[m],
            (
                floor((w - wm) / 2),
                ceil((w - wm) / 2),
                floor((h - hm) / 2),
                ceil((h - hm) / 2),
            ),
            mode="constant",
            value=0,
        )

    return data, (3, h, w)



from .base import BaseSampler, BaseSamplerConfig
from .gaussian_mixture import GaussianMixtureSampler, GaussianMixtureSamplerConfig
from .iaf_sampler import IAFSampler, IAFSamplerConfig
from .maf_sampler import MAFSampler, MAFSamplerConfig

__all__ = [
    "BaseSampler",
    "BaseSamplerConfig",
    "GaussianMixtureSampler",
    "GaussianMixtureSamplerConfig",
    "IAFSamplerConfig",
    "IAFSampler",
    "MAFSampler",
    "MAFSamplerConfig",
]



from typing import List

import torch
from torch import Tensor


class cca_loss:
    def __init__(self, outdim_size, use_all_singular_values):
        self.outdim_size = outdim_size
        self.use_all_singular_values = use_all_singular_values

    def loss(self, H_list: List[Tensor]):
        """

        It is the loss function of CCA as introduced in the original paper. There can be other formulations.

        """

        r1 = 1e-3
        r2 = 1e-3
        eps = 1e-9
        H1, H2 = H_list[0], H_list[1]
        H1, H2 = H1.t(), H2.t()
        device = H1.device
        # assert torch.isnan(H1).sum().item() == 0
        # assert torch.isnan(H2).sum().item() == 0

        o1 = o2 = H1.size(0)

        m = H1.size(1)
        #         print(H1.size())

        H1bar = H1 - H1.mean(dim=1).unsqueeze(dim=1)
        H2bar = H2 - H2.mean(dim=1).unsqueeze(dim=1)
        # assert torch.isnan(H1bar).sum().item() == 0
        # assert torch.isnan(H2bar).sum().item() == 0

        SigmaHat12 = (1.0 / (m - 1)) * torch.matmul(H1bar, H2bar.t())
        SigmaHat11 = (1.0 / (m - 1)) * torch.matmul(H1bar, H1bar.t()) + r1 * torch.eye(
            o1
        ).to(device)
        SigmaHat22 = (1.0 / (m - 1)) * torch.matmul(H2bar, H2bar.t()) + r2 * torch.eye(
            o2
        ).to(device)
        # assert torch.isnan(SigmaHat11).sum().item() == 0
        # assert torch.isnan(SigmaHat12).sum().item() == 0
        # assert torch.isnan(SigmaHat22).sum().item() == 0

        # Calculating the root inverse of covariance matrices by using eigen decomposition
        # [D1, V1] = torch.symeig(SigmaHat11, eigenvectors=True)
        # [D2, V2] = torch.symeig(SigmaHat22, eigenvectors=True)
        [D1, V1] = torch.linalg.eigh(SigmaHat11, UPLO="U")
        [D2, V2] = torch.linalg.eigh(SigmaHat22, UPLO="U")
        # assert torch.isnan(D1).sum().item() == 0
        # assert torch.isnan(D2).sum().item() == 0
        # assert torch.isnan(V1).sum().item() == 0
        # assert torch.isnan(V2).sum().item() == 0

        # Added to increase stability
        posInd1 = torch.gt(D1, eps).nonzero()[:, 0]
        D1 = D1[posInd1]
        V1 = V1[:, posInd1]
        posInd2 = torch.gt(D2, eps).nonzero()[:, 0]
        D2 = D2[posInd2]
        V2 = V2[:, posInd2]
        # print(posInd1.size())
        # print(posInd2.size())

        SigmaHat11RootInv = torch.matmul(
            torch.matmul(V1, torch.diag(D1**-0.5).to(device)), V1.t()
        )
        SigmaHat22RootInv = torch.matmul(
            torch.matmul(V2, torch.diag(D2**-0.5).to(device)), V2.t()
        )

        Tval = torch.matmul(
            torch.matmul(SigmaHat11RootInv, SigmaHat12), SigmaHat22RootInv
        )
        #         print(Tval.size())

        if self.use_all_singular_values:
            # all singular values are used to calculate the correlation
            tmp = torch.matmul(Tval.t(), Tval)
            corr = torch.trace(torch.sqrt(tmp))
            # assert torch.isnan(corr).item() == 0
        else:
            # just the top self.outdim_size singular values are used
            trace_TT = torch.matmul(Tval.t(), Tval)
            trace_TT = torch.add(
                trace_TT, (torch.eye(trace_TT.shape[0]).to(device) * r1)
            )  # regularization for more stability
            U = torch.linalg.eigvalsh(trace_TT, UPLO="U")
            U = torch.where(U > eps, U, (torch.ones(U.shape).double().to(device) * eps))
            U = U.topk(self.outdim_size)[0]
            corr = torch.sum(torch.sqrt(U))
        return -corr


class mcca_loss:

    """
    Wraps-up several cca loss when training with more than two modalities.
    """

    def __init__(self, outdim_size, use_all_singular_values) -> None:
        self.outdim_size = outdim_size
        self.use_all_singular_values = use_all_singular_values
        self.cca_loss = cca_loss(outdim_size, use_all_singular_values).loss

    def loss(self, H_list):
        loss = 0
        for i, h1 in enumerate(H_list):
            for j, h2 in enumerate(H_list):
                if i < j:
                    loss += self.cca_loss([h1, h2])

        return loss



import os

import pytest
import torch
import torch.optim as optim
from pythae.models.base import BaseAEConfig
from pythae.models.nn.benchmarks.mnist.convnets import (
    Decoder_Conv_AE_MNIST,
    Encoder_Conv_VAE_MNIST,
)
from pythae.models.nn.default_architectures import Encoder_VAE_MLP

from multivae.data import MultimodalBaseDataset
from multivae.models import JMVAE, JMVAEConfig
from multivae.models.nn.default_architectures import Decoder_AE_MLP
from multivae.trainers import BaseTrainer, BaseTrainerConfig
from multivae.trainers.base.callbacks import rename_logs

PATH = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def training_config(tmpdir):
    tmpdir.mkdir("dummy_folder")
    dir_path = os.path.join(tmpdir, "dummy_folder")
    return BaseTrainerConfig(output_dir=dir_path)


@pytest.fixture
def model_sample():
    model_config = JMVAEConfig(n_modalities=2, latent_dim=10)
    config = BaseAEConfig(input_dim=(1, 28, 28), latent_dim=10)
    encoders = dict(mod1=Encoder_VAE_MLP(config), mod2=Encoder_Conv_VAE_MNIST(config))
    decoders = dict(mod1=Decoder_AE_MLP(config), mod2=Decoder_Conv_AE_MNIST(config))
    return JMVAE(model_config=model_config, encoders=encoders, decoders=decoders)


@pytest.fixture
def train_dataset():
    return MultimodalBaseDataset(
        data=dict(
            mod1=torch.randn((2, 1, 28, 28)),
            mod2=torch.randn((2, 1, 28, 28)),
        ),
        labels=torch.tensor([0, 1]),
    )


class Test_Set_Training_config:
    @pytest.fixture(
        params=[
            None,
            BaseTrainerConfig(),
            BaseTrainerConfig(
                per_device_train_batch_size=10,
                per_device_eval_batch_size=20,
                learning_rate=1e-5,
                optimizer_cls="AdamW",
                optimizer_params={"weight_decay": 0.01},
                scheduler_cls="ExponentialLR",
                scheduler_params={"gamma": 0.321},
            ),
        ]
    )
    def training_configs(self, request, tmpdir):
        if request.param is not None:
            tmpdir.mkdir("dummy_folder")
            dir_path = os.path.join(tmpdir, "dummy_folder")
            request.param.output_dir = dir_path
            return request.param
        else:
            return None

    def test_set_training_config(self, model_sample, train_dataset, training_configs):
        trainer = BaseTrainer(
            model=model_sample,
            train_dataset=train_dataset,
            training_config=training_configs,
        )

        # check if default config is set
        if training_configs is None:
            assert trainer.training_config == BaseTrainerConfig(
                output_dir="dummy_output_dir", keep_best_on_train=True
            )

        else:
            assert trainer.training_config == training_configs


class Test_Build_Optimizer:
    def test_wrong_optimizer_cls(self):
        with pytest.raises(AttributeError):
            BaseTrainerConfig(optimizer_cls="WrongOptim")

    def test_wrong_optimizer_params(self):
        with pytest.raises(TypeError):
            BaseTrainerConfig(
                optimizer_cls="Adam", optimizer_params={"wrong_config": 1}
            )

    @pytest.fixture(params=[BaseTrainerConfig(), BaseTrainerConfig(learning_rate=1e-5)])
    def training_configs_learning_rate(self, tmpdir, request):
        request.param.output_dir = tmpdir.mkdir("dummy_folder")
        return request.param

    @pytest.fixture(
        params=[
            {"optimizer_cls": "Adagrad", "optimizer_params": {"lr_decay": 0.1}},
            {"optimizer_cls": "AdamW", "optimizer_params": {"betas": (0.1234, 0.4321)}},
            {"optimizer_cls": "SGD", "optimizer_params": None},
        ]
    )
    def optimizer_config(self, request, training_configs_learning_rate):
        optimizer_config = request.param

        # set optim and params to training config
        training_configs_learning_rate.optimizer_cls = optimizer_config["optimizer_cls"]
        training_configs_learning_rate.optimizer_params = optimizer_config[
            "optimizer_params"
        ]

        return optimizer_config

    def test_default_optimizer_building(
        self, model_sample, train_dataset, training_configs_learning_rate
    ):
        trainer = BaseTrainer(
            model=model_sample,
            train_dataset=train_dataset,
            training_config=training_configs_learning_rate,
        )

        trainer.set_optimizer()

        assert issubclass(type(trainer.optimizer), torch.optim.Adam)
        assert (
            trainer.optimizer.defaults["lr"]
            == training_configs_learning_rate.learning_rate
        )

    def test_set_custom_optimizer(
        self,
        model_sample,
        train_dataset,
        training_configs_learning_rate,
        optimizer_config,
    ):
        trainer = BaseTrainer(
            model=model_sample,
            train_dataset=train_dataset,
            training_config=training_configs_learning_rate,
        )

        trainer.set_optimizer()

        assert issubclass(
            type(trainer.optimizer),
            getattr(torch.optim, optimizer_config["optimizer_cls"]),
        )
        assert (
            trainer.optimizer.defaults["lr"]
            == training_configs_learning_rate.learning_rate
        )
        if optimizer_config["optimizer_params"] is not None:
            assert all(
                [
                    trainer.optimizer.defaults[key]
                    == optimizer_config["optimizer_params"][key]
                    for key in optimizer_config["optimizer_params"].keys()
                ]
            )


class Test_Build_Scheduler:
    def test_wrong_scheduler_cls(self):
        with pytest.raises(AttributeError):
            BaseTrainerConfig(scheduler_cls="WrongOptim")

    def test_wrong_scheduler_params(self):
        with pytest.raises(TypeError):
            BaseTrainerConfig(
                scheduler_cls="ReduceLROnPlateau", scheduler_params={"wrong_config": 1}
            )

    @pytest.fixture(params=[BaseTrainerConfig(), BaseTrainerConfig(learning_rate=1e-5)])
    def training_configs_learning_rate(self, tmpdir, request):
        request.param.output_dir = tmpdir.mkdir("dummy_folder")
        return request.param

    @pytest.fixture(
        params=[
            {"optimizer_cls": "Adagrad", "optimizer_params": {"lr_decay": 0.1}},
            {"optimizer_cls": "AdamW", "optimizer_params": {"betas": (0.1234, 0.4321)}},
            {"optimizer_cls": "SGD", "optimizer_params": None},
        ]
    )
    def optimizer_config(self, request, training_configs_learning_rate):
        optimizer_config = request.param

        # set optim and params to training config
        training_configs_learning_rate.optimizer_cls = optimizer_config["optimizer_cls"]
        training_configs_learning_rate.optimizer_params = optimizer_config[
            "optimizer_params"
        ]

        return optimizer_config

    @pytest.fixture(
        params=[
            {"scheduler_cls": "StepLR", "scheduler_params": {"step_size": 1}},
            {"scheduler_cls": "LinearLR", "scheduler_params": None},
            {"scheduler_cls": "ExponentialLR", "scheduler_params": {"gamma": 3.14}},
        ]
    )
    def scheduler_config(self, request, training_configs_learning_rate):
        scheduler_config = request.param

        # set scheduler and params to training config
        training_configs_learning_rate.scheduler_cls = scheduler_config["scheduler_cls"]
        training_configs_learning_rate.scheduler_params = scheduler_config[
            "scheduler_params"
        ]

        return request.param

    def test_default_scheduler_building(
        self, model_sample, train_dataset, training_configs_learning_rate
    ):
        trainer = BaseTrainer(
            model=model_sample,
            train_dataset=train_dataset,
            training_config=training_configs_learning_rate,
        )

        trainer.set_optimizer()
        trainer.set_scheduler()

        assert trainer.scheduler is None

    def test_set_custom_scheduler(
        self,
        model_sample,
        train_dataset,
        training_configs_learning_rate,
        scheduler_config,
    ):
        trainer = BaseTrainer(
            model=model_sample,
            train_dataset=train_dataset,
            training_config=training_configs_learning_rate,
        )

        trainer.set_optimizer()
        trainer.set_scheduler()

        assert issubclass(
            type(trainer.scheduler),
            getattr(torch.optim.lr_scheduler, scheduler_config["scheduler_cls"]),
        )
        if scheduler_config["scheduler_params"] is not None:
            assert all(
                [
                    trainer.scheduler.state_dict()[key]
                    == scheduler_config["scheduler_params"][key]
                    for key in scheduler_config["scheduler_params"].keys()
                ]
            )


class Test_Device_Checks:
    def test_set_environ_variable(self):
        os.environ["LOCAL_RANK"] = "1"
        os.environ["WORLD_SIZE"] = "4"
        os.environ["RANK"] = "3"
        os.environ["MASTER_ADDR"] = "314"
        os.environ["MASTER_PORT"] = "222"

        trainer_config = BaseTrainerConfig()

        assert int(trainer_config.local_rank) == 1
        assert int(trainer_config.world_size) == 4
        assert int(trainer_config.rank) == 3
        assert trainer_config.master_addr == "314"
        assert trainer_config.master_port == "222"

        del os.environ["LOCAL_RANK"]
        del os.environ["WORLD_SIZE"]
        del os.environ["RANK"]
        del os.environ["MASTER_ADDR"]
        del os.environ["MASTER_PORT"]

    @pytest.fixture(
        params=[
            BaseTrainerConfig(num_epochs=3, no_cuda=True),
        ]
    )
    def training_configs(self, tmpdir, request):
        tmpdir.mkdir("dummy_folder")
        dir_path = os.path.join(tmpdir, "dummy_folder")
        request.param.output_dir = dir_path
        return request.param

    def test_setup_device_with_no_cuda(
        self, model_sample, train_dataset, training_configs
    ):
        trainer = BaseTrainer(
            model=model_sample,
            train_dataset=train_dataset,
            eval_dataset=train_dataset,
            training_config=training_configs,
        )

        device = trainer._setup_devices()
        assert device == "cpu"


class TestPredict:
    def test_default_optimizer_building(
        self, model_sample, train_dataset, training_config
    ):
        trainer = BaseTrainer(
            model=model_sample,
            train_dataset=train_dataset,
            eval_dataset=train_dataset,
            training_config=training_config,
        )

        all_recons = trainer.predict(model_sample, epoch=1, n_data=3)

        assert list(all_recons.keys()) == model_sample.modalities_name


class TestSaving:
    @pytest.fixture(
        params=[
            BaseTrainerConfig(num_epochs=3, no_cuda=True),
        ]
    )
    def training_configs(self, tmpdir, request):
        dir_path = os.path.join(tmpdir, "test_output_dir")
        request.param.output_dir = dir_path
        return request.param

    def test_create_dir(self, tmpdir, model_sample, train_dataset, training_configs):
        trainer = BaseTrainer(
            model=model_sample,
            train_dataset=train_dataset,
            eval_dataset=train_dataset,
            training_config=training_configs,
        )

        assert os.path.exists(os.path.join(tmpdir, "test_output_dir"))


class TestLogging:
    @pytest.fixture
    def log_output_dir(self):
        return "dummy_log_output_dir"

    def test_create_dir(
        self, tmpdir, model_sample, train_dataset, training_config, log_output_dir
    ):
        trainer = BaseTrainer(
            model=model_sample,
            train_dataset=train_dataset,
            eval_dataset=train_dataset,
            training_config=training_config,
        )

        # create dummy training signature
        trainer._training_signature = "dummy_signature"

        assert not os.path.exists(os.path.join(tmpdir, "dummy_log_output_dir"))
        file_logger = trainer._get_file_logger(os.path.join(tmpdir, log_output_dir))

        assert os.path.exists(os.path.join(tmpdir, "dummy_log_output_dir"))
        assert os.path.exists(
            os.path.join(
                tmpdir, "dummy_log_output_dir", f"training_logs_dummy_signature.log"
            )
        )


class TestTrainingCallbacks:
    def test_rename_logs(self):
        dummy_metrics = {"train_metric": 12, "eval_metric": 13}

        renamed_metrics = rename_logs(dummy_metrics)

        assert set(renamed_metrics.keys()).issubset(
            set(["train/metric", "eval/metric"])
        )



from dataclasses import field
from typing import Dict, Literal, Tuple, Union

from pydantic.dataclasses import dataclass
from pythae.config import BaseConfig


@dataclass
class BaseMultiVAEConfig(BaseConfig):
    """This is the base config for a Multi-Modal VAE model.

    Parameters:
        n_modalities (int): The number of modalities. Default: None.
        latent_dim (int): The dimension of the latent space. Default: None.
        input_dims (dict[str,tuple]) : The modalities'names (str) and input shapes (tuple).
        uses_likelihood_rescaling (bool): To mitigate modality collapse, it is possible to use likelihood rescaling.
            (see : https://proceedings.mlr.press/v162/javaloy22a.html).
            The inputs_dim must be provided to compute the likelihoods rescalings. It is used in a number of models
            which is why we include it here. Default to False.
        rescale_factors (dict[str, float]): The reconstruction rescaling factors per modality.
            If None is provided but uses_likelihood_rescaling is True, a default value proportional to the input modality
            size is computed. Default to None.
        decoders_dist (Dict[str, Union[function, str]]). The decoder distributions to use per modality.
            Per modality, you can provide a string in ['normal','bernoulli','laplace']. For Bernoulli distribution,
            the decoder is expected to output **logits**. If None is provided, a normal distribution is used for each modality.
        decoder_dist_params (Dict[str,dict]) : Parameters for the output decoder distributions, for
            computing the log-probability.
            For instance, with normal or laplace distribution, you can pass the scale in this dictionary.
            ex :  {'mod1' : {scale : 0.75}}
    """

    n_modalities: Union[int, None] = None
    latent_dim: int = 10
    input_dims: dict = None
    uses_likelihood_rescaling: bool = False
    rescale_factors: dict = None
    decoders_dist: Dict[
        str, Literal["normal", "bernoulli", "laplace", "categorical"]
    ] = None
    decoder_dist_params: dict = None
    custom_architectures: list = field(default_factory=lambda: [])


@dataclass
class EnvironmentConfig(BaseConfig):
    python_version: str = "3.8"


@dataclass
class BaseAEConfig(BaseConfig):
    """This is the base configuration instance of encoders/decoders models deriving from
    :class:`~pythae.config.BaseConfig`.

    Parameters:
        input_dim (tuple): The input_data dimension (channels X x_dim X y_dim)
        latent_dim (int): The latent space dimension. Default: None.
        style_dim (int) : For models with private latent spaces for each modality. Default: 0.
    """

    input_dim: Union[Tuple[int, ...], None] = None
    latent_dim: int = 10
    style_dim: int = 0



import logging
from typing import List, Optional

from multivae.data.datasets.base import MultimodalBaseDataset
from multivae.models.base.base_model import BaseMultiVAE
from multivae.trainers.base.callbacks import TrainingCallback

from ..base import BaseTrainer
from .jnf_trainer_config import TwoStepsTrainerConfig

logger = logging.getLogger(__name__)

# make it print to the console.
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class TwoStepsTrainer(BaseTrainer):
    """A specific trainer that handles the training of the joint VAE models.

    Args:
        model (BaseMultiVAE): A instance of :class:`~multivae.models.BaseMultiVAE` to train

        train_dataset (MultimodalBaseDataset): The training dataset of type
            :class:`~multivae.data.datasets.MultimodalBaseDataset`

        eval_dataset (MultimodalBaseDataset): The evaluation dataset of type
            :class:`~multivae.data.datasets.MultimodalBaseDataset`

        training_config (BaseTrainerConfig): The training arguments summarizing the main
            parameters used for training. If None, a basic training instance of
            :class:`BaseTrainerConfig` is used. Default: None.

        callbacks (List[~pythae.trainers.training_callbacks.TrainingCallback]):
            A list of callbacks to use during training.
    """

    def __init__(
        self,
        model: BaseMultiVAE,
        train_dataset: MultimodalBaseDataset,
        eval_dataset: Optional[MultimodalBaseDataset] = None,
        training_config: Optional[TwoStepsTrainerConfig] = None,
        callbacks: List[TrainingCallback] = None,
    ):
        super().__init__(model, train_dataset, eval_dataset, training_config, callbacks)

    def prepare_train_step(self, epoch, best_train_loss, best_eval_loss):
        """
        Function to operate changes between train_steps such as resetting the optimizer and
        the best losses values.
        """
        if epoch in self.model.reset_optimizer_epochs:
            logger.info(f"Epoch {epoch} : reset the optimizer and losses.")
            # Reset the optimizer
            self.set_optimizer()
            self.set_scheduler()
            best_train_loss = 1e12
            best_eval_loss = 1e12
        return best_train_loss, best_eval_loss
